# %% imports
import autogaita.gui.gaita_widgets as gaita_widgets
import autogaita.gui.gui_utils as gui_utils
from autogaita.gui.first_level_gui_utils import (
    update_config_file,
    extract_cfg_from_json_file,
    extract_results_from_json_file,
)
from autogaita.resources.utils import try_to_run_gaita
from autogaita.universal3D.universal3D_datafile_preparation import prepare_3D
import tkinter as tk
import customtkinter as ctk
import os
from threading import Thread
import platform


# %% global constants
from autogaita.gui.gui_constants import (
    UNIVERSAL3D_FG_COLOR,
    UNIVERSAL3D_HOVER_COLOR,
    HEADER_FONT_NAME,
    HEADER_FONT_SIZE,
    MAIN_HEADER_FONT_SIZE,
    TEXT_FONT_NAME,
    TEXT_FONT_SIZE,
    ADV_CFG_TEXT_FONT_SIZE,
    CLOSE_COLOR,
    CLOSE_HOVER_COLOR,
    COLOR_PALETTES_LIST,
    WINDOWS_TASKBAR_MAXHEIGHT,
    AUTOGAITA_FOLDER_PATH,
    get_widget_cfg_dict,  # function!
)

# these colors are GUI-specific - add to common widget cfg
FG_COLOR = UNIVERSAL3D_FG_COLOR
HOVER_COLOR = UNIVERSAL3D_HOVER_COLOR
WIDGET_CFG = get_widget_cfg_dict()
WIDGET_CFG["FG_COLOR"] = FG_COLOR
WIDGET_CFG["HOVER_COLOR"] = HOVER_COLOR

CONFIG_FILE_NAME = "universal3D_gui_config.json"
INT_VARS = ["sampling_rate", "bin_num", "plot_joint_number"]
LIST_VARS = ["joints", "y_standardisation_joint", "z_standardisation_joint"]
DICT_VARS = ["angles"]
# TK_BOOL/STR_VARS are only used for initialising widgets based on cfg file
# (note that numbers are initialised as strings)
TK_BOOL_VARS = [
    "analyse_singlerun",
    "dont_show_plots",
    "y_acceleration",
    "angular_acceleration",
    "plot_SE",
    "standardise_z_at_SC_level",
    "standardise_z_to_a_joint",
    "standardise_y_coordinates",
    "postname_flag",
    "flip_gait_direction",
    "analyse_average_y",
    "legend_outside",
    "fileprep_3D_DLC",
]
TK_STR_VARS = [
    "sampling_rate",
    "bin_num",
    "plot_joint_number",
    "results_dir",
    "name",
    "root_dir",
    "sctable_filename",
    "postname_string",
    "color_palette",
    "coordinate_standardisation_xls",
    "fileprep_root_dir",  # for datafile prep window only
    "fileprep_filetype",
    "fileprep_postname_string",
    "fileprep_results_dir",
    "fileprep_separator",
    "fileprep_string_to_remove",
]


# %% An important Note
# I am using a global variable called cfg because I need its info to be shared
# between root and columnconfiguration window. This is not the object-oriented way
# that one would do this typically. However, it works as expected since:
# 1) cfg's values are only ever modified @ initialisation & by widgets
# 2) cfg's values are shared for analysis of a single and multiple video(s)
# 3) cfg's values are passed to all functions that need them
# 4) and (IMPORTANTLY!) just before running either (i.e. single/multi) analysis, cfg's
#    and result's values are unpacked and assigned to "this_" dicts that are passed
#    to the runanalysis local function. Hence, from that point onwards, only
#    "this_" dicts are used, never cfg or result dicts themselves.
#    ==> see donewindow for point 4)

# %%............................  MAIN PROGRAM ................................


def run_universal3D_gui():
    # ..........................................................................
    # ......................  root window initialisation .......................
    # ..........................................................................
    # Check for config file
    config_file_path = os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME)
    if not os.path.isfile(config_file_path):
        config_file_error_msg = (
            "universal3D_gui_config.json file not found in autogaita folder.\n"
            "Confirm that the file exists and is named correctly.\n"
            "If not, download it again from the GitHub repository."
        )
        tk.messagebox.showerror(
            title="Config File Error", message=config_file_error_msg
        )
        exit()

    # CustomTkinter vars
    ctk.set_appearance_mode("dark")  # Modes: system (default), light, dark
    ctk.set_default_color_theme("green")  # Themes: blue , dark-blue, green
    # root
    root = ctk.CTk()
    # make window pretty
    screen_width = root.winfo_screenwidth()  # width of the screen
    screen_height = root.winfo_screenheight()  # height of the screen
    if platform.system() == "Windows":  # adjust for taskbar in windows only
        screen_height -= WINDOWS_TASKBAR_MAXHEIGHT
    # create root dimensions (based on fullscreen) to pass to other window-functions l8r
    w, h, x, y = screen_width, screen_height, 0, 0
    root_dimensions = (w, h, x, y)
    # set the dimensions of the screen and where it is placed
    # => have it half-wide starting at 1/4 of screen's width (dont change w & x!)
    root.geometry(f"{int(screen_width / 2)}x{screen_height}+{int(screen_width / 4)}+0")
    root.title("Universal 3D GaitA")
    gui_utils.fix_window_after_its_creation(root)
    gui_utils.configure_the_icon(root)

    # ..........................................................................
    # ........................  root window population .........................
    # ..........................................................................
    # initialise default cfg vars
    # ==> see a note @ get_results_and_cfg helper function about why I
    #     initialise all vars that are not Boolean as Strings!
    # ==> note that cfg is global because we need to change some of its variables
    #     column configure window
    # ==> the results dict is not global and will be passed from function to function
    #     like proper programmers do it (nice)
    global cfg  # global cfg variable

    cfg = extract_cfg_from_json_file(
        root,
        AUTOGAITA_FOLDER_PATH,
        CONFIG_FILE_NAME,
        LIST_VARS,
        DICT_VARS,
        TK_STR_VARS,
        TK_BOOL_VARS,
    )
    results = extract_results_from_json_file(
        root, AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME, TK_STR_VARS, TK_BOOL_VARS
    )

    # .........................  main configuration  ...................................
    # config header
    cfgheader_label = gaita_widgets.header_label(root, "Main Configuration", WIDGET_CFG)
    cfgheader_label.grid(row=0, column=0, columnspan=2, sticky="nsew")

    # root directory
    rootdir_label, rootdir_entry = gaita_widgets.label_and_entry_pair(
        root,
        "Directory containing the files to be analysed:",
        results["root_dir"],
        WIDGET_CFG,
    )
    rootdir_label.grid(row=1, column=0, columnspan=2, sticky="w")
    rootdir_entry.grid(row=1, column=1)

    # stepcycle latency XLS
    SCXLS_label, SCXLS_entry = gaita_widgets.label_and_entry_pair(
        root,
        "Name of the Annotation Table Excel file:",
        results["sctable_filename"],
        WIDGET_CFG,
    )
    SCXLS_label.grid(row=2, column=0, columnspan=2, sticky="w")
    SCXLS_entry.grid(row=2, column=1)

    # sampling rate
    samprate_label, samprate_entry = gaita_widgets.label_and_entry_pair(
        root,
        "Sampling rate of videos in Hertz (frames/second):",
        cfg["sampling_rate"],
        WIDGET_CFG,
    )
    samprate_label.grid(row=3, column=0, columnspan=2, sticky="w")
    samprate_entry.grid(row=3, column=1)

    # flip gait direction
    flip_gait_direction_box = gaita_widgets.checkbox(
        root,
        "Standardise direction of movement across runs",
        cfg["flip_gait_direction"],
        WIDGET_CFG,
    )
    flip_gait_direction_box.grid(row=4, column=0, columnspan=2, sticky="w")

    # postname present checkbox
    postname_flag_checkbox = gaita_widgets.checkbox(
        root,
        "I have a constant string that follows IDs in filenames.",
        results["postname_flag"],
        WIDGET_CFG,
    )
    postname_flag_checkbox.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            results, "postname_flag", postname_entry
        ),
    )
    postname_flag_checkbox.grid(row=5, column=0, columnspan=2)

    # postname string
    postname_label, postname_entry = gaita_widgets.label_and_entry_pair(
        root,
        "The constant string is:",
        results["postname_string"],
        WIDGET_CFG,
    )
    postname_label.grid(row=6, column=0, sticky="e")
    postname_entry.configure(state="disabled")
    postname_entry.grid(row=6, column=1, sticky="w")

    # empty label 1 (for spacing)
    empty_label_one = ctk.CTkLabel(root, text="")
    empty_label_one.grid(row=7, column=0)

    # ..........................  advanced cfg section  ................................

    # advanced header string
    advanced_cfg_header_label = gaita_widgets.header_label(
        root, "Advanced Configuration", WIDGET_CFG
    )
    advanced_cfg_header_label.grid(row=8, column=0, columnspan=2, sticky="nsew")

    # data file preparation window
    datafile_prep_button = gaita_widgets.header_button(
        root, "Data File Preparation", WIDGET_CFG
    )
    datafile_prep_button.configure(
        command=lambda: build_datafile_prep_window(root, results, cfg),
    )
    datafile_prep_button.grid(row=9, column=0, columnspan=2)

    # column name information window
    column_info_button = gaita_widgets.header_button(
        root, "Customise Joints & Angles", WIDGET_CFG
    )
    column_info_button.configure(
        command=lambda: build_column_info_window(root, cfg, root_dimensions),
    )
    column_info_button.grid(row=10, column=0, columnspan=2)

    # advanced cfg
    cfg_window_button = gaita_widgets.header_button(
        root, "Advanced Configuration", WIDGET_CFG
    )
    cfg_window_button.configure(
        command=lambda: (build_cfg_window(root, cfg)),
    )
    cfg_window_button.grid(row=11, column=0, columnspan=2)

    # empty label 2 (for spacing)
    empty_label_two = ctk.CTkLabel(root, text="")
    empty_label_two.grid(row=12, column=0)

    # ............................  run & exit section  ................................

    # run analysis label
    runheader_label = gaita_widgets.header_label(root, "Run Analysis", WIDGET_CFG)
    runheader_label.grid(row=13, column=0, columnspan=2, sticky="nsew")

    # single video checkbox
    singlevideo_checkbox = gaita_widgets.checkbox(
        root,
        "Only analyse a single dataset.",
        cfg["analyse_singlerun"],
        WIDGET_CFG,
    )
    singlevideo_checkbox.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "analyse_singlerun", ID_entry
        ),
    )
    singlevideo_checkbox.grid(row=14, column=0, columnspan=2)

    # ID string info
    ID_label, ID_entry = gaita_widgets.label_and_entry_pair(
        root, "ID to be analysed:", results["name"], WIDGET_CFG
    )
    ID_label.grid(row=15, column=0, sticky="e")
    ID_entry.grid(row=15, column=1, sticky="w")
    gui_utils.change_widget_state_based_on_checkbox(cfg, "analyse_singlerun", ID_entry)

    # run analysis button
    run_button = gaita_widgets.header_button(root, "Run Analysis!", WIDGET_CFG)
    run_button.configure(
        command=lambda: (
            update_config_file(
                results,
                cfg,
                AUTOGAITA_FOLDER_PATH,
                CONFIG_FILE_NAME,
                LIST_VARS,
                DICT_VARS,
                TK_STR_VARS,
                TK_BOOL_VARS,
            ),
            build_donewindow(results, root, root_dimensions),
        )
    )
    run_button.grid(row=16, column=0, columnspan=2, padx=10, pady=(10, 5))

    # empty label 3 (for spacing)
    empty_label_three = ctk.CTkLabel(root, text="")
    empty_label_three.grid(row=17, column=0)

    # close program button
    exit_button = gaita_widgets.exit_button(
        root,
        WIDGET_CFG,
    )
    exit_button.configure(
        command=lambda: (
            update_config_file(
                results,
                cfg,
                AUTOGAITA_FOLDER_PATH,
                CONFIG_FILE_NAME,
                LIST_VARS,
                DICT_VARS,
                TK_STR_VARS,
                TK_BOOL_VARS,
            ),
            root.withdraw(),
            root.after(5000, root.destroy),
        ),
    )
    exit_button.grid(row=18, column=0, columnspan=2, padx=10, pady=(10, 5))

    # maximise widgets
    gui_utils.maximise_widgets(root)
    root.columnconfigure(list(range(2)), weight=1, uniform="Silent_Creme")

    # main loop
    root.mainloop()


# %%..............  LOCAL FUNCTION(S) #1 - BUILD DATAFILE PREP WINDOW  .................
def build_datafile_prep_window(root, results, cfg):
    """Build data file preparation window"""
    # unpack root window dimensions
    screen_width = root.winfo_screenwidth()  # width of the screen
    screen_height = root.winfo_screenheight()  # height of the screen
    # build window
    fileprep_window = ctk.CTkToplevel(root)
    fileprep_window.title("Datafile Preparation")
    fileprep_window.geometry(
        f"{int(screen_width / 2)}x{int(screen_height / 1.5)}+{int(screen_width / 4)}+{int(screen_height / 6)}"
    )
    gui_utils.fix_window_after_its_creation(fileprep_window)

    # fileprep header
    fileprep_header_label = gaita_widgets.header_label(
        fileprep_window, "Datafile Preparation", WIDGET_CFG
    )
    fileprep_header_label.grid(row=0, column=0, rowspan=2, columnspan=2, sticky="nsew")
    # root dir
    fileprep_root_dir_label, fileprep_root_dir_entry = (
        gaita_widgets.label_and_entry_pair(
            fileprep_window,
            "Directory containing files:",
            cfg["fileprep_root_dir"],
            WIDGET_CFG,
        )
    )
    fileprep_root_dir_label.grid(row=2, column=0, sticky="e")
    fileprep_root_dir_entry.grid(row=2, column=1)
    # file type
    filetype_string = "File type:"
    filetype_label = ctk.CTkLabel(
        fileprep_window,
        text=filetype_string,
        font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
    )
    filetype_label.grid(row=3, column=0, columnspan=2)
    filetype_strings = ["csv", "xls", "xlsx"]
    filetype_buttons = []
    for i, string in enumerate(filetype_strings):
        filetype_buttons.append(
            ctk.CTkRadioButton(
                fileprep_window,
                text=string,
                variable=cfg["fileprep_filetype"],
                value=string,
                fg_color=FG_COLOR,
                hover_color=HOVER_COLOR,
                font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
            )
        )
        filetype_buttons[i].grid(row=4 + i, column=0, columnspan=2)
    row_num = 4 + len(filetype_strings)  # 7
    # postname string
    fileprep_postname_label, fileprep_postname_entry = (
        gaita_widgets.label_and_entry_pair(
            fileprep_window,
            "Constant string present in file names:",
            cfg["fileprep_postname_string"],
            WIDGET_CFG,
        )
    )
    fileprep_postname_label.grid(row=row_num, column=0, sticky="e")
    fileprep_postname_entry.grid(row=row_num, column=1)
    # results dir
    fileprep_results_dir_label, fileprep_results_dir_entry = (
        gaita_widgets.label_and_entry_pair(
            fileprep_window,
            "Directory to save results to:",
            cfg["fileprep_results_dir"],
            WIDGET_CFG,
        )
    )
    fileprep_results_dir_label.grid(row=row_num + 1, column=0, sticky="e")
    fileprep_results_dir_entry.grid(row=row_num + 1, column=1)
    # empty label for spacing
    empty_label_one = ctk.CTkLabel(fileprep_window, text="")
    empty_label_one.grid(row=row_num + 2, column=0)

    # ............................  left section: clean  ...............................
    # cleaning header
    clean_header_label = gaita_widgets.header_label(
        fileprep_window, "Clean Columns", WIDGET_CFG
    )
    clean_header_label.grid(row=row_num + 3, column=0, sticky="nsew")
    # string to remove
    string_to_remove_label, string_to_remove_entry = gaita_widgets.label_and_entry_pair(
        fileprep_window,
        "Text to remove from all columns:",
        cfg["fileprep_string_to_remove"],
        WIDGET_CFG,
    )
    string_to_remove_label.grid(row=row_num + 4, column=0)
    string_to_remove_entry.grid(row=row_num + 5, column=0)
    # cleaning run button
    clean_button = gaita_widgets.header_button(fileprep_window, "Clean!", WIDGET_CFG)
    clean_button.configure(
        command=lambda: (
            update_config_file(
                results,
                cfg,
                AUTOGAITA_FOLDER_PATH,
                CONFIG_FILE_NAME,
                LIST_VARS,
                DICT_VARS,
                TK_STR_VARS,
                TK_BOOL_VARS,
            ),
            run_universal_3D_preparation("clean", cfg),
        )
    )
    clean_button.grid(row=row_num + 7, column=0, sticky="nsew", padx=10, pady=(10, 5))

    # ............................  right section: rename  .............................
    # renaming header
    rename_header_label = gaita_widgets.header_label(
        fileprep_window, "Rename Columns", WIDGET_CFG
    )
    rename_header_label.grid(row=row_num + 3, column=1, sticky="nsew")
    # separator
    separator_label, separator_entry = gaita_widgets.label_and_entry_pair(
        fileprep_window,
        "Separator (must be unique!):",
        cfg["fileprep_separator"],
        WIDGET_CFG,
    )
    separator_label.grid(row=row_num + 4, column=1)
    separator_entry.grid(row=row_num + 5, column=1)
    # 3D DLC checkbox
    DLC_checkbox = gaita_widgets.checkbox(
        fileprep_window,
        "3D DeepLabCut files",
        cfg["fileprep_3D_DLC"],
        WIDGET_CFG,
    )
    DLC_checkbox.grid(row=row_num + 6, column=1)
    # renaming run button
    rename_button = gaita_widgets.header_button(fileprep_window, "Rename!", WIDGET_CFG)
    rename_button.configure(
        command=lambda: (
            update_config_file(
                results,
                cfg,
                AUTOGAITA_FOLDER_PATH,
                CONFIG_FILE_NAME,
                LIST_VARS,
                DICT_VARS,
                TK_STR_VARS,
                TK_BOOL_VARS,
            ),
            run_universal_3D_preparation("rename", cfg),
        )
    )
    rename_button.grid(row=row_num + 7, column=1, sticky="nsew", padx=10, pady=(10, 5))

    # ............................  bottom section: done  ..............................
    # empty label two
    empty_label_two = ctk.CTkLabel(fileprep_window, text="")
    empty_label_two.grid(row=row_num + 7, column=0)
    # done button
    done_button = ctk.CTkButton(
        fileprep_window,
        text="I am done!",
        fg_color=CLOSE_COLOR,
        hover_color=CLOSE_HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: fileprep_window.destroy(),
    )
    done_button.grid(
        row=row_num + 8, column=0, columnspan=2, padx=30, pady=(10, 5), sticky="ew"
    )
    # maximise widgets
    gui_utils.maximise_widgets(fileprep_window)


def run_universal_3D_preparation(task, cfg):
    """Call the autogaita_3D_preparation function with appropriate arguments"""
    # create separate cfg dict for fileprep function to avoid changing the main cfg
    fileprep_cfg = {}
    fileprep_cfg["root_dir"] = cfg["fileprep_root_dir"].get()
    fileprep_cfg["file_type"] = cfg["fileprep_filetype"].get()
    fileprep_cfg["postname_string"] = cfg["fileprep_postname_string"].get()
    fileprep_cfg["results_dir"] = cfg["fileprep_results_dir"].get()
    fileprep_cfg["3D_DLC"] = cfg["fileprep_3D_DLC"].get()
    if fileprep_cfg["3D_DLC"] is True:
        task = "DLC"
    # set the appropriate kwargs based on the task
    if task == "clean":
        string_to_remove = cfg["fileprep_string_to_remove"].get()
        kwargs = {"string_to_remove": string_to_remove}
    elif task in ["rename", "DLC"]:
        separator = cfg["fileprep_separator"].get()
        kwargs = {"separator": separator}
    # run the function, with kwargs depending on task, & show info for users
    info_message = prepare_3D(task, fileprep_cfg, **kwargs)
    tk.messagebox.showinfo(title="Info", message=info_message)


# %%..........  LOCAL FUNCTION(S) #2 - BUILD ADVANCED CFG WINDOW  .............


def build_cfg_window(root, cfg):
    """Build advanced configuration window"""
    # unpack root window dimensions
    screen_width = root.winfo_screenwidth()  # width of the screen
    screen_height = root.winfo_screenheight()  # height of the screen
    # build window
    cfg_window = ctk.CTkToplevel(root)
    cfg_window.title("Advanced Configuration")
    cfg_window.geometry(
        f"{int(screen_width / 2)}x{screen_height}+{int(screen_width / 4)}+0"
    )
    gui_utils.fix_window_after_its_creation(cfg_window)

    #  ...........................  advanced analysis  .................................
    # advanced analysis header
    adv_cfg_analysis_header_label = gaita_widgets.header_label(
        cfg_window, "Analysis", WIDGET_CFG
    )
    adv_cfg_analysis_header_label.grid(
        row=0, column=0, rowspan=2, columnspan=2, sticky="nsew"
    )
    # bin_num
    bin_num_label, bin_num_entry = gaita_widgets.label_and_entry_pair(
        cfg_window,
        "Number of bins to use for normalising the step cycle:",
        cfg["bin_num"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    bin_num_label.grid(row=2, column=0, columnspan=2)
    bin_num_entry.grid(row=3, column=0, columnspan=2)
    # acceleration label
    acceleration_string = "Analyse (i.e. plot & export) accelerations for:"
    acceleration_label = ctk.CTkLabel(
        cfg_window,
        text=acceleration_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    acceleration_label.grid(row=4, column=0, columnspan=2)
    # y acceleration
    y_accel_box = gaita_widgets.checkbox(
        cfg_window,
        "y-coordinates",
        cfg["y_acceleration"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    y_accel_box.grid(row=5, column=0, sticky="e")
    # angular acceleration
    angular_accel_box = gaita_widgets.checkbox(
        cfg_window,
        "angles",
        cfg["angular_acceleration"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    angular_accel_box.grid(row=5, column=1, sticky="w")
    # height standardisation at each step cycle separately
    z_standardisation_box = gaita_widgets.checkbox(
        cfg_window,
        "Normalise heights separately for all step cycles",
        cfg["standardise_z_at_SC_level"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    z_standardisation_box.grid(row=6, column=0, columnspan=2)
    # z standardisation to a specific joint not global minimum
    standardise_z_to_joint_box = gaita_widgets.checkbox(
        cfg_window,
        "Standardise heights to a joint instead of to global minimum",
        cfg["standardise_z_to_a_joint"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    standardise_z_to_joint_box.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "standardise_z_to_a_joint", z_standardisation_joint_entry
        ),
    )
    standardise_z_to_joint_box.grid(row=7, column=0, columnspan=2)
    # z standardisation joint string label & entry
    z_standardisation_joint_label, z_standardisation_joint_entry = (
        gaita_widgets.label_and_entry_pair(
            cfg_window,
            "Height-standardisation joint:",
            cfg["z_standardisation_joint"][0],
            WIDGET_CFG,
            adv_cfg_textsize=True,
        )
    )
    z_standardisation_joint_label.grid(row=8, column=0, sticky="e")
    z_standardisation_joint_entry.grid(row=8, column=1, sticky="w")
    # to initialise the widget correctly, run this function once
    gui_utils.change_widget_state_based_on_checkbox(
        cfg, "standardise_z_to_a_joint", z_standardisation_joint_entry
    )
    # export average y coordinates
    analyse_average_y_box = gaita_widgets.checkbox(
        cfg_window,
        "Analyse y-coordinate averages",
        cfg["analyse_average_y"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    analyse_average_y_box.grid(row=9, column=0, columnspan=2)
    analyse_average_y_box.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "analyse_average_y", standardise_y_coordinates_box
        ),
    )
    # standardise y coordinates
    standardise_y_coordinates_box = gaita_widgets.checkbox(
        cfg_window,
        "Standardise y-coordinates",
        cfg["standardise_y_coordinates"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    standardise_y_coordinates_box.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "standardise_y_coordinates", y_standardisation_joint_entry
        ),
    )
    standardise_y_coordinates_box.grid(row=10, column=0, columnspan=2)
    gui_utils.change_widget_state_based_on_checkbox(
        cfg, "analyse_average_y", standardise_y_coordinates_box
    )
    # y standardisation joint string label & entry
    y_standardisation_joint_label, y_standardisation_joint_entry = (
        gaita_widgets.label_and_entry_pair(
            cfg_window,
            "Y-standardisation joint:",
            cfg["y_standardisation_joint"][0],
            WIDGET_CFG,
            adv_cfg_textsize=True,
        )
    )
    y_standardisation_joint_label.grid(row=11, column=0, sticky="e")
    y_standardisation_joint_entry.grid(row=11, column=1, sticky="w")
    gui_utils.change_widget_state_based_on_checkbox(
        cfg, "standardise_y_coordinates", y_standardisation_joint_entry
    )
    # standardise all (primary) joint coordinates by a fixed decimal value
    coordinate_standardisation_xls_label, coordinate_standardisation_xls_entry = (
        gaita_widgets.label_and_entry_pair(
            cfg_window,
            "Excel file for primary-joint coordinate standardisation:",
            cfg["coordinate_standardisation_xls"],
            WIDGET_CFG,
            adv_cfg_textsize=True,
        )
    )
    coordinate_standardisation_xls_label.grid(row=12, column=0, columnspan=2)
    coordinate_standardisation_xls_entry.grid(row=13, column=0, columnspan=2)

    #  .............................  advanced output  .................................
    # advanced output header
    adv_cfg_output_header_label = gaita_widgets.header_label(
        cfg_window, "Output", WIDGET_CFG
    )
    adv_cfg_output_header_label.grid(
        row=14, column=0, rowspan=2, columnspan=2, sticky="nsew"
    )
    # number of joints to plot
    plot_joint_num_label, plot_joint_num_entry = gaita_widgets.label_and_entry_pair(
        cfg_window,
        "Number of joints to plot in detail:",
        cfg["plot_joint_number"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    plot_joint_num_label.grid(row=16, column=0, sticky="e")
    plot_joint_num_entry.grid(row=16, column=1, sticky="w")
    # plot plots to python
    showplots_checkbox = gaita_widgets.checkbox(
        cfg_window,
        "Don't show plots in GUI (save only)",
        cfg["dont_show_plots"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    showplots_checkbox.grid(row=17, column=0, columnspan=2)
    # plot SE
    plot_SE_box = gaita_widgets.checkbox(
        cfg_window,
        "Use standard error instead of standard deviation for plots",
        cfg["plot_SE"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    plot_SE_box.grid(row=18, column=0, columnspan=2)
    # color palette
    color_palette_string = "Choose figures' color palette"
    color_palette_label = ctk.CTkLabel(
        cfg_window,
        text=color_palette_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    color_palette_label.grid(row=19, column=0, columnspan=2)
    color_palette_entry = ctk.CTkOptionMenu(
        cfg_window,
        values=COLOR_PALETTES_LIST,
        variable=cfg["color_palette"],
        fg_color=FG_COLOR,
        button_color=FG_COLOR,
        button_hover_color=HOVER_COLOR,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    color_palette_entry.grid(row=20, column=0, columnspan=2)
    # legend outside
    legend_outside_checkbox = gaita_widgets.checkbox(
        cfg_window,
        "Plot legends outside of figures' panels",
        cfg["legend_outside"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    legend_outside_checkbox.grid(row=21, column=0, columnspan=2)

    # results dir
    results_dir_label, results_dir_entry = gaita_widgets.label_and_entry_pair(
        cfg_window,
        "Save Results subfolders to directory below instead of to data's",
        cfg["results_dir"],
        WIDGET_CFG,
        adv_cfg_textsize=True,
    )
    results_dir_label.grid(row=22, column=0, columnspan=2)
    results_dir_entry.grid(row=23, column=0, columnspan=2)

    # done button
    adv_cfg_done_button = ctk.CTkButton(
        cfg_window,
        text="I am done, update cfg.",
        fg_color=CLOSE_COLOR,
        hover_color=CLOSE_HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: cfg_window.destroy(),
    )
    adv_cfg_done_button.grid(
        row=24, column=0, rowspan=2, columnspan=2, sticky="nsew", padx=10, pady=(10, 5)
    )
    # maximise widgets
    cfg_window.columnconfigure(list(range(2)), weight=1, uniform="Silent_Creme")
    cfg_window.rowconfigure(list(range(20)), weight=1, uniform="Silent_Creme")


# %%..............  LOCAL FUNCTION(S) #3 - BUILD COLUMN INFO WINDOW  ...................
def build_column_info_window(root, cfg, root_dimensions):
    """Build a window allowing users to configure custom column names if required"""
    columnwindow = ctk.CTkToplevel(root)
    columnwindow.geometry("%dx%d+%d+%d" % root_dimensions)
    columnwindow.title("Custom column names & features")
    gui_utils.fix_window_after_its_creation(columnwindow)

    # .............  Nested Function: Add joint label & entry  .........................
    def add_joint(window, key):
        """Add a joint if needed or a angle name/lower_joint/upper_joint if angle widget"""
        # append StringVar to cfg appropriately
        if key == "angles":
            for angle_key in cfg[key]:
                cfg[key][angle_key].append(tk.StringVar(root, ""))
        else:
            cfg[key].append(tk.StringVar(root, ""))
        # find out the number of rows to append to it
        nrows = window.grid_size()[1]
        # add stuff based on to which case we are adding
        if key == "joints":
            label, entry = gaita_widgets.label_and_entry_pair(
                window,
                "Joint #" + str(len(cfg[key])),
                cfg[key][-1],
                WIDGET_CFG,
            )
            label.grid(row=nrows + 1, column=0, sticky="ew")
            entry.grid(row=nrows + 2, column=0)
        elif key == "angles":
            for a, angle_key in enumerate(cfg[key]):
                if angle_key == "name":
                    this_case = "Angle"
                elif angle_key == "lower_joint":
                    this_case = "Lower Joint"
                elif angle_key == "upper_joint":
                    this_case = "Upper Joint"
                label, entry = gaita_widgets.label_and_entry_pair(
                    window,
                    this_case + " #" + str(len(cfg[key][angle_key])),
                    cfg[key][angle_key][-1],
                    WIDGET_CFG,
                )
                label.grid(row=nrows + 1, column=angle_column + a, sticky="ew")
                entry.grid(row=nrows + 2, column=angle_column + a)
        # maximise columns
        for c in range(window.grid_size()[0]):
            window.grid_columnconfigure(c, weight=1)

    # ...................  Scrollable Window Configuration  ............................
    scrollable_rows = 8

    # ...................  Column 0: hind limb joint names  ............................
    joint_column = 0
    # header label
    joint_label = gaita_widgets.header_label(columnwindow, "Joints", WIDGET_CFG)
    joint_label.grid(row=0, column=joint_column, sticky="nsew", pady=(0, 5))
    # initialise scrollable frame for joints
    joint_frame = ctk.CTkScrollableFrame(columnwindow)
    joint_frame.grid(row=1, column=joint_column, rowspan=scrollable_rows, sticky="nsew")
    # initialise labels & entries with joint defaults
    initialise_labels_and_entries(joint_frame, "joints", "Joint ")
    # add joint button
    add_joint_button = gaita_widgets.header_button(
        columnwindow, "Add joint", WIDGET_CFG
    )
    add_joint_button.configure(
        command=lambda: (add_joint(joint_frame, "joints")),  # 2nd input = cfg's key
    )
    add_joint_button.grid(
        row=2 + scrollable_rows,
        column=joint_column,
        sticky="nsew",
        padx=5,
        pady=(10, 5),
    )

    # .........  Column 1: angle names/joint-definitions & done button  ................
    angle_column = 1
    # header label
    angle_label = gaita_widgets.header_label(
        columnwindow, "Angle Configuration", WIDGET_CFG
    )
    angle_label.grid(
        row=0, column=angle_column, columnspan=3, sticky="nsew", pady=(0, 5)
    )
    # initialise scrollable frame for angles
    angle_frame = ctk.CTkScrollableFrame(columnwindow)
    angle_frame.grid(
        row=1, column=angle_column, rowspan=scrollable_rows, columnspan=3, sticky="nsew"
    )
    # initial entries & labels
    for a, angle_key in enumerate(cfg["angles"]):
        if angle_key == "name":
            this_case = "Angle"
        elif angle_key == "lower_joint":
            this_case = "Lower Joint"
        elif angle_key == "upper_joint":
            this_case = "Upper Joint"
        initialise_labels_and_entries(
            angle_frame, ["angles", angle_key], this_case, angle_column + a
        )
    # add angle trio button
    add_angle_button = gaita_widgets.header_button(
        columnwindow, "Add angle", WIDGET_CFG
    )
    add_angle_button.configure(
        command=lambda: add_joint(angle_frame, "angles"),  # 2nd input = cfg's key
    )
    add_angle_button.grid(
        row=2 + scrollable_rows,
        column=angle_column,
        columnspan=3,
        sticky="nsew",
        padx=5,
        pady=(10, 5),
    )
    # done button
    columncfg_done_button = ctk.CTkButton(
        columnwindow,
        text="I am done, update cfg!",
        fg_color=CLOSE_COLOR,
        hover_color=CLOSE_HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: (columnwindow.destroy()),
    )
    columncfg_done_button.grid(
        row=3 + scrollable_rows,
        column=joint_column,
        columnspan=4,
        sticky="nsew",
        padx=200,
        pady=10,
    )
    # maximise everything in columnwindow
    gui_utils.maximise_widgets(columnwindow)


def initialise_labels_and_entries(window, key, which_case_string, *args):
    """Add labels & entries for joint column information areas"""
    # we input a list of strings if we use this function to initialise angles (since
    # cfg["angle"] itself is a dict) => handle this particularly at textvariable
    # parameter of CTkEntry
    if type(key) is str:
        this_var = cfg[key]
    elif type(key) is list:
        this_var = cfg[key[0]][key[1]]
    joint_labels = [[] for _ in this_var]
    joint_entries = [[] for _ in this_var]
    row_counter = 0  # because we always call this in a scrollable frame
    if args:
        column_number = args
    else:
        column_number = 0
    for j, joint in enumerate(this_var):
        joint_labels[j] = ctk.CTkLabel(
            window,
            text=which_case_string + " #" + str(j + 1),
            font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
        )
        joint_labels[j].grid(row=row_counter, column=column_number, sticky="ew")
        row_counter += 1
        if type(key) is str:
            joint_entries[j] = ctk.CTkEntry(
                window, textvariable=cfg[key][j], font=(TEXT_FONT_NAME, TEXT_FONT_SIZE)
            )
        elif type(key) is list:
            joint_entries[j] = ctk.CTkEntry(
                window,
                textvariable=cfg[key[0]][key[1]][j],
                font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
            )
        # span only in width
        joint_entries[j].grid(row=row_counter, column=column_number)
        row_counter += 1
    # maximise columns
    for c in range(window.grid_size()[0]):
        window.grid_columnconfigure(c, weight=1)


# %%..................  LOCAL FUNCTION(S) #4 - BUILD DONE WINDOW  ......................
def build_donewindow(results, root, root_dimensions):
    """Build done window informing people"""

    # .........................................................................
    # .............. IMPORTANT - GET VARS & CHECK USER-INPUT ..................
    # .........................................................................
    # ==> Extract "this" results & cfg info IMMEDIATELY BEFORE running analysis
    #     (i.e., before providing donewindow)
    # ==> get_results_and_cfg checks whether the numerical vars (see
    #     FLOAT_/INT_VARS) were numbers - if not it returns an error_msg, which
    #     we use to inform users about their wrong input
    #       -- Catch this here and don't even show donewindow if input is wrong
    # ==> NEVER, EVER change the global variable cfg!
    try:
        this_runs_results, this_runs_cfg = get_results_and_cfg(results, cfg)
    except:
        error_msg = get_results_and_cfg(results, cfg)
        tk.messagebox.showerror(title="Try again", message=error_msg)
        return

    # create the window & make it pretty and nice
    w = root_dimensions[0]
    h = root_dimensions[1]
    donewindow = ctk.CTkToplevel(root)
    donewindow.title("GaitA Ready :)")
    donewindow_w = w * (3 / 5)
    donewindow_h = h * (1 / 5)
    donewindow_x = (w - donewindow_w) // 2
    donewindow_y = h * (1 / 2.5)
    donewindow.geometry(
        "%dx%d+%d+%d" % (donewindow_w, donewindow_h, donewindow_x, donewindow_y)
    )
    gui_utils.fix_window_after_its_creation(donewindow)

    # labels
    done_label1_string = "Your results will be saved as your data is processed"
    done_label1 = ctk.CTkLabel(
        donewindow,
        text=done_label1_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label1.grid(row=0, column=0, sticky="nsew")
    done_label2_string = (
        "Please see the Python command window for progress "
        + "and the figure panel for an overview of all plots."
    )
    done_label2 = ctk.CTkLabel(
        donewindow,
        text=done_label2_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label2.grid(row=1, column=0, sticky="nsew")
    done_label3_string = (
        "You may start another analysis while we are "
        + "processing - however your PC might slow down a bit. "
    )
    done_label3 = ctk.CTkLabel(
        donewindow,
        text=done_label3_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label3.grid(row=2, column=0, sticky="nsew")
    done_label4_string = (
        "Thank you for using AutoGaitA! Feel free to "
        + "contact me on Github or at autogaita@fz-juelich.de - MH."
    )
    done_label4 = ctk.CTkLabel(
        donewindow,
        text=done_label4_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label4.grid(row=3, column=0, sticky="nsew")
    # run button
    done_button = ctk.CTkButton(
        donewindow,
        text="Run!",
        fg_color=FG_COLOR,
        hover_color=HOVER_COLOR,
        font=(HEADER_FONT_NAME, MAIN_HEADER_FONT_SIZE),
        command=lambda: (
            run_analysis(this_runs_results, this_runs_cfg),
            donewindow.destroy(),
        ),
    )
    done_button.grid(row=4, column=0, sticky="nsew", pady=10, padx=200)

    gui_utils.maximise_widgets(donewindow)


# %%..........  LOCAL FUNCTION(S) #5 - PREPARE & CALL AUTOGAITA  ....................


def run_analysis(this_runs_results, this_runs_cfg):
    """Run the main program"""
    if this_runs_cfg["analyse_singlerun"]:
        run_thread = Thread(
            target=analyse_single_run, args=(this_runs_results, this_runs_cfg)
        )
    else:
        run_thread = Thread(
            target=analyse_multi_run, args=(this_runs_results, this_runs_cfg)
        )
    run_thread.start()


def analyse_single_run(this_runs_results, this_runs_cfg):
    """Prepare for one execution of autogaita_universal3D & execute"""
    # prepare
    this_info = {}  # info dict: run-specific info
    this_info["name"] = this_runs_results["name"]
    this_folderinfo = prepare_folderinfo(this_runs_results)
    if this_folderinfo is None:
        error_msg = (
            "No directory found at: " + this_runs_results["root_dir"] + " - try again!"
        )
        tk.messagebox.showerror(title="Try again", message=error_msg)
        print(error_msg)
        return
    if this_runs_cfg["results_dir"]:
        this_info["results_dir"] = os.path.join(
            this_runs_cfg["results_dir"], this_info["name"]
        )
    else:
        this_info["results_dir"] = os.path.join(
            this_runs_results["root_dir"], "Results", this_info["name"]
        )
    # execute
    try_to_run_gaita("Universal 3D", this_info, this_folderinfo, this_runs_cfg, False)


def analyse_multi_run(this_runs_results, this_runs_cfg):
    """Analyse multiple runs at once - extracts folderinfo and then loops over all names
    of multirun_info
    """
    this_folderinfo = prepare_folderinfo(this_runs_results)
    if this_folderinfo is None:
        error_msg = (
            "No directory found at: " + this_runs_results["root_dir"] + " - try again!"
        )
        tk.messagebox.showerror(title="Try again", message=error_msg)
        print(error_msg)
        return
    multirun_info = multirun_extract_info(this_folderinfo)
    for idx, name in enumerate(multirun_info["name"]):
        multirun_run_a_single_dataset(
            idx, multirun_info, this_folderinfo, this_runs_cfg
        )


def multirun_run_a_single_dataset(idx, multirun_info, this_folderinfo, this_runs_cfg):
    """If we are doing a multi-run analysis, run the main code of individual analyses
    based on current cfg"""
    # extract and pass info of this ID
    this_info = {}
    keynames = multirun_info.keys()
    for keyname in keynames:
        this_info[keyname] = multirun_info[keyname][idx]
    # important to only pass this_info here (1 run at a time - prints error if needed)
    try_to_run_gaita("Universal 3D", this_info, this_folderinfo, this_runs_cfg, True)


# %%..............  LOCAL FUNCTION(S) #6 - VARIOUS HELPER FUNCTIONS  ...................


def get_results_and_cfg(results, cfg):
    """Before calling analysis, use .get() to extract values from tk-vars"""
    # 1) We make sure to return and use for all "run"-purposes (i.e. everything
    #    that follows this local function) "this_" dicts (stuff is dangerous
    #    otherwise!)
    # 2) I implemented the usage & transformation of FLOAT_&INT_VARS the way I
    #    do here (instead of initialising as (e.g.) tk.IntVar because otherwise
    #    tkinter complains whenever an Entry is empty @ any time, just because
    #    the user is inputting their values.
    output_dicts = [{}, {}]
    for i in range(len(output_dicts)):
        if i == 0:
            input_dict = results
        elif i == 1:
            input_dict = cfg
        for key in input_dict.keys():
            if key in INT_VARS:
                try:
                    output_dicts[i][key] = int(input_dict[key].get())
                except ValueError:
                    return (
                        key
                        + " value was not an integer number, but "
                        + input_dict[key].get()
                    )
            elif key in LIST_VARS:
                # if list of strings, initialise output empty list and get & append vals
                output_dicts[i][key] = []
                for list_idx in range(len(input_dict[key])):
                    output_dicts[i][key].append(input_dict[key][list_idx].get())
            elif key in DICT_VARS:
                # if dict of list of strings, initialise as empty dict and assign stuff
                # key = "angles" or other DICT_VAR
                # inner_key = "name" & "lower_ / upper_joint"
                # list_idx = idx of list of strings of inner_key
                output_dicts[i][key] = {}
                for inner_key in input_dict[key]:
                    output_dicts[i][key][inner_key] = []
                    for list_idx in range(len(input_dict[key][inner_key])):
                        output_dicts[i][key][inner_key].append(
                            input_dict[key][inner_key][list_idx].get()
                        )
            else:
                output_dicts[i][key] = input_dict[key].get()
    this_runs_results = output_dicts[0]
    this_runs_cfg = output_dicts[1]
    return this_runs_results, this_runs_cfg


def prepare_folderinfo(this_runs_results):
    """Dump all infos about constants in this given folder into a dict"""
    # IMPORTANT NOTE
    # This function should only return None if there are issues with root_dir
    # since we handle the error in outside functions because of this (i.e. the meaning
    # behind the None should NOT change in the future!)
    if len(this_runs_results["root_dir"]) == 0:
        return
    folderinfo = {}
    folderinfo["root_dir"] = this_runs_results["root_dir"]
    # to make sure root_dir works under windows
    # (windows is okay with all dir-separators being "/", so make sure it is!)
    folderinfo["root_dir"] = folderinfo["root_dir"].replace(os.sep, "/")
    if folderinfo["root_dir"][-1] != "/":
        folderinfo["root_dir"] = folderinfo["root_dir"] + "/"
    if not os.path.exists(folderinfo["root_dir"]):
        return
    folderinfo["sctable_filename"] = this_runs_results["sctable_filename"]
    if this_runs_results["postname_flag"] is False:
        folderinfo["postname_string"] = ""
    else:
        folderinfo["postname_string"] = this_runs_results["postname_string"]
    return folderinfo


def multirun_extract_info(folderinfo):
    """If we are running multi-run analysis, prepare a dict of lists that include
    unique name & results_dir infos

    A Note
    ------
    There are 3 "results_dirs" here:
    1) cfg["results_dir"] is the user-provided entry
    2) results_dir var is the value extracted from 1)
    3) info["results_dir"] is the dir we save this ID's (!) Results to!
    """

    root_dir = folderinfo["root_dir"]
    results_dir = cfg["results_dir"].get()  # global var....
    info = {"name": [], "results_dir": []}
    for filename in os.listdir(folderinfo["root_dir"]):
        # dont try to combine the two "join" if blocks into one - we want to append
        # results dir WHENEVER we append name!
        if not folderinfo["postname_string"]:
            # dont use endswith below to catch .xlsx too
            if (".xls" in filename) & (folderinfo["sctable_filename"] not in filename):
                info["name"].append(filename.split(".xls")[0])
                if results_dir:
                    info["results_dir"].append(
                        os.path.join(results_dir, info["name"][-1])
                    )
                else:
                    info["results_dir"].append(
                        os.path.join(
                            folderinfo["root_dir"], "Results", info["name"][-1]
                        )
                    )
        else:
            if folderinfo["postname_string"] in filename:
                info["name"].append(filename.split(folderinfo["postname_string"])[0])
                if results_dir:
                    info["results_dir"].append(
                        os.path.join(results_dir, info["name"][-1])
                    )
                else:
                    info["results_dir"].append(
                        os.path.join(
                            folderinfo["root_dir"], "Results", info["name"][-1]
                        )
                    )
    if len(info["name"]) < 1:
        no_files_message = (
            f"Unable to find any files at {root_dir}!" + "\ncheck your inputs!"
        )
        tk.messagebox.showerror(
            title="No files found!",
            message=no_files_message,
        )
    return info


# %% what happens if we hit run
if __name__ == "__main__":
    run_universal3D_gui()



from autogaita.common2D.common2D_1_preparation import (
    some_prep,
    move_data_to_folders,
    check_gait_direction,
)
import os
import pandas as pd
import pytest


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_data_using_some_prep(extract_info, extract_folderinfo, extract_cfg):
    data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    return data


@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["mouse_num"] = 15
    info["run_num"] = 3
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/dlc_data"
    folderinfo["sctable_filename"] = (
        "correct_annotation_table.xlsx"  # has to be an excel file
    )
    folderinfo["data_string"] = "SIMINewOct"
    folderinfo["beam_string"] = "BeamTraining"
    folderinfo["premouse_string"] = "Mouse"
    folderinfo["postmouse_string"] = "25mm"
    folderinfo["prerun_string"] = "run"
    folderinfo["postrun_string"] = "6DLC"
    return folderinfo


@pytest.fixture
def extract_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = True
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = True
    cfg["pixel_to_mm_ratio"] = 3.76
    cfg["x_sc_broken_threshold"] = 200
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = True
    cfg["y_standardisation_joint"] = ["Knee"]
    cfg["plot_joint_number"] = 3
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_x"] = True
    cfg["standardise_x_coordinates"] = True
    cfg["x_standardisation_joint"] = ["Hind paw tao"]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["hind_joints"] = ["Hind paw tao", "Ankle", "Knee", "Hip", "Iliac Crest"]
    cfg["fore_joints"] = [
        "Front paw tao ",
        "Wrist ",
        "Elbow ",
        "Lower Shoulder ",
        "Upper Shoulder ",
    ]
    cfg["beam_col_left"] = ["BeamLeft"]  # BEAM_COL_LEFT & _RIGHT must be lists of len=1
    cfg["beam_col_right"] = ["BeamRight"]
    cfg["beam_hind_jointadd"] = ["Tail base ", "Tail center ", "Tail tip "]
    cfg["beam_fore_jointadd"] = ["Nose ", "Ear base "]
    cfg["angles"] = {
        "name": ["Ankle ", "Knee ", "Hip "],
        "lower_joint": ["Hind paw tao ", "Ankle ", "Knee "],
        "upper_joint": ["Knee ", "Hip ", "Iliac Crest "],
    }
    return cfg


# %%..............................  preparation  .......................................


def test_move_csv_datafile_to_results_dir(extract_info, extract_folderinfo):
    move_data_to_folders("DLC", ".csv", extract_info, extract_folderinfo)
    for file in os.listdir(extract_info["results_dir"]):
        if file.endswith(".csv"):
            assert (
                (extract_folderinfo["premouse_string"] in file)
                & (extract_folderinfo["postmouse_string"] in file)
                & (extract_folderinfo["prerun_string"] in file)
                & (extract_folderinfo["postrun_string"] in file)
            )


def test_check_gait_direction(extract_data_using_some_prep, extract_cfg, extract_info):
    direction_joint = extract_cfg["direction_joint"]
    flip_gait_direction = True  # 1) test broken DLC data (empty)
    broken_data = pd.DataFrame(data=None, columns=extract_data_using_some_prep.columns)
    check_gait_direction(
        "DLC", broken_data, direction_joint, flip_gait_direction, extract_info
    )
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Unable to determine gait direction!" in content
    flip_this_data = pd.DataFrame(
        data=None, columns=extract_data_using_some_prep.columns
    )  # 2) test data that has to be flipped
    data_len = 20
    flip_this_data[direction_joint + "likelihood"] = [0.99] * data_len
    first_half = [10] * int((data_len / 2))
    second_half = [-10] * int((data_len / 2))
    x_coords = first_half + second_half
    flip_this_data[direction_joint + "x"] = x_coords
    flip_this_data = check_gait_direction(
        "DLC", flip_this_data, direction_joint, flip_gait_direction, extract_info
    )
    assert flip_this_data["Flipped"][0] == True



from autogaita.common2D.common2D_1_preparation import (
    check_and_expand_cfg,
    check_and_fix_cfg_strings,
    flip_mouse_body,
    some_prep,  # note that first input of some_prep is set to "DLC" when not mattering!
)
import os
import copy
import math
import numpy as np
import pandas.testing as pdt
from hypothesis import given, strategies as st, settings, HealthCheck
import pytest


# %%..............................  fixtures  ..........................................
@pytest.fixture
def extract_data_using_some_prep(extract_info, extract_folderinfo, extract_cfg):
    data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    return data


@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["mouse_num"] = 15
    info["run_num"] = 3
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/dlc_data"
    folderinfo["sctable_filename"] = (
        "correct_annotation_table.xlsx"  # has to be an excel file
    )
    folderinfo["data_string"] = "SIMINewOct"
    folderinfo["beam_string"] = "BeamTraining"
    folderinfo["premouse_string"] = "Mouse"
    folderinfo["postmouse_string"] = "25mm"
    folderinfo["prerun_string"] = "run"
    folderinfo["postrun_string"] = "6DLC"
    return folderinfo


@pytest.fixture
def extract_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = True
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = True
    cfg["pixel_to_mm_ratio"] = 3.76
    cfg["x_sc_broken_threshold"] = 200
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = True
    cfg["y_standardisation_joint"] = ["Knee"]
    cfg["plot_joint_number"] = 3
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_x"] = True
    cfg["standardise_x_coordinates"] = True
    cfg["x_standardisation_joint"] = ["Hind paw tao"]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["hind_joints"] = ["Hind paw tao", "Ankle", "Knee", "Hip", "Iliac Crest"]
    cfg["fore_joints"] = [
        "Front paw tao ",
        "Wrist ",
        "Elbow ",
        "Lower Shoulder ",
        "Upper Shoulder ",
    ]
    cfg["beam_col_left"] = ["BeamLeft"]  # BEAM_COL_LEFT & _RIGHT must be lists of len=1
    cfg["beam_col_right"] = ["BeamRight"]
    cfg["beam_hind_jointadd"] = ["Tail base ", "Tail center ", "Tail tip "]
    cfg["beam_fore_jointadd"] = ["Nose ", "Ear base "]
    cfg["angles"] = {
        "name": ["Ankle ", "Knee ", "Hip "],
        "lower_joint": ["Hind paw tao ", "Ankle ", "Knee "],
        "upper_joint": ["Knee ", "Hip ", "Iliac Crest "],
    }
    return cfg


# %%..............................  preparation  .......................................
#                       AN IMPORTANT NOTE ON THESE UNIT TESTS!
# Calling check_and_expand cfg outside of some_prep with the
# export_data_using_some_prep fixture leads to the cfg that is returned by
# check_and_expand to be None since the data var DOES NOT INCLUDE the beam!
# => We thus set cfg["subtract_beam"] to False prior to calling it (see e.g.the
#    plot_joint test)


# %%..........................  cfg & string stuff  ....................................
def test_plot_joint_error(extract_data_using_some_prep, extract_cfg, extract_info):
    extract_cfg["plot_joint_number"] = 2000
    extract_cfg["subtract_beam"] = False
    check_and_expand_cfg(extract_data_using_some_prep, extract_cfg, extract_info)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "we can :)" in content
    extract_cfg = check_and_expand_cfg(
        extract_data_using_some_prep, extract_cfg, extract_info
    )
    assert extract_cfg["plot_joints"] == extract_cfg["hind_joints"]
    extract_cfg["plot_joint_number"] = 2
    extract_cfg = check_and_expand_cfg(
        extract_data_using_some_prep, extract_cfg, extract_info
    )
    assert extract_cfg["plot_joints"] == extract_cfg["hind_joints"][:2]


def test_error_if_no_cfgkey_joints(extract_info, extract_folderinfo, extract_cfg):
    full_cfg = copy.deepcopy(extract_cfg)  # no referencing here - we need copies!
    for cfg_key in [
        "hind_joints",
        "x_standardisation_joint",
        "y_standardisation_joint",
    ]:
        extract_cfg = copy.deepcopy(full_cfg)  # here too!
        extract_cfg[cfg_key] = ["not_in_data"]
        data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
        with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
            content = f.read()
        if cfg_key == "hind_joints":
            assert "hind limb joint names" in content
        elif cfg_key == "x_standardisation_joint":
            assert "x-coordinate standardisation joint" in content
        elif cfg_key == "y_standardisation_joint":
            assert "y-coordinate standardisation joint" in content
        assert data is None
    # cannot loop case of x & y joints being broken
    extract_cfg = copy.deepcopy(full_cfg)
    extract_cfg["x_standardisation_joint"] = ["not_in_data"]
    extract_cfg["y_standardisation_joint"] = ["not_in_data"]
    data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
        assert "x & y-coordinate standardisation joint" in content
        assert data is None


@given(test_list=st.lists(st.text(), min_size=1))
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_removal_of_wrong_strings_from_cfg_key(
    test_list, extract_data_using_some_prep, extract_cfg, extract_info
):
    # the following loop is to account for hypothesis randomly generating strings that
    # actually are data columns (happend for "Knee ")
    for i in range(len(test_list)):
        if test_list[i] + "x" in extract_data_using_some_prep.columns:
            test_list.pop(i)
    cfg_key = "hind_joints"  # irrelevant since property testing
    extract_cfg[cfg_key] = test_list
    test_result = check_and_fix_cfg_strings(
        extract_data_using_some_prep, extract_cfg, cfg_key, extract_info
    )
    assert not test_result  # empty list is falsey


def test_wrong_data_and_beam_strings(extract_info, extract_folderinfo, extract_cfg):
    extract_folderinfo["beam_string"] = extract_folderinfo["data_string"]
    some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Your data & baseline (beam) identifiers ([G] in our" in content


def test_wrong_postmouse_string(extract_info, extract_folderinfo, extract_cfg):
    extract_folderinfo["postmouse_string"] = "this_is_a_test"
    some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Unable to identify ANY RELEVANT FILES for" in content


# %%...........................  dataframe stuff  ......................................
def test_cols_we_added_to_data(extract_info, extract_folderinfo, extract_cfg):
    data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    assert (data.columns[0] == "Time") & (data.columns[1] == "Flipped")


def test_datas_indexing_and_time_column(extract_info, extract_folderinfo, extract_cfg):
    for extract_cfg["sampling_rate"] in [50, 500, 5000]:
        data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
        # use isclose here because there are some floating point things going on (eg. 1.
        # 66 and 1.660 for sampling rate of 500)
        assert math.isclose(
            data["Time"].max(),
            (len(data) - 1) / (1 * extract_cfg["sampling_rate"]),
            rel_tol=1e-9,
        )


# %%...........................  data manipulation  ....................................
def test_global_min_standardisation(extract_info, extract_folderinfo, extract_cfg):
    extract_cfg["subtract_beam"] = False
    extract_cfg["standardise_y_to_a_joint"] = False
    data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    y_cols = [c for c in data.columns if c.endswith(" y")]
    assert data[y_cols].min().min() == 0
    # approach here is find difference between global & standardisation joint minma and
    # see if all y cols' difference is equal to that
    # => this implies that joint-based y-standardisation worked
    global_and_standardisation_joints_y_min_diff = data[
        extract_cfg["y_standardisation_joint"][0] + " y"
    ].min()
    global_min_data = data.copy()
    extract_cfg["standardise_y_to_a_joint"] = True
    data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    assert np.allclose(  # use np.allclose here because we are comparing arrays
        global_min_data[y_cols],
        data[y_cols] + global_and_standardisation_joints_y_min_diff,
        atol=1e-9,
    )


def test_flip_mouse_body(extract_info, extract_folderinfo, extract_cfg):
    extract_cfg["flip_gait_direction"] = False
    test_data = some_prep("DLC", extract_info, extract_folderinfo, extract_cfg)
    function_flipped_data = test_data.copy()
    function_flipped_data = flip_mouse_body(test_data, extract_info)
    x_cols = [col for col in function_flipped_data.columns if col.endswith(" x")]
    global_x_max = max(test_data[x_cols].max())
    for col in x_cols:
        function_flipped_series = function_flipped_data[col]
        function_flipped_series = function_flipped_series.astype(float)
        manually_flipped_series = global_x_max - test_data[col]
        pdt.assert_series_equal(function_flipped_series, manually_flipped_series)



# %% imports
import autogaita.gui.gaita_widgets as gaita_widgets
import autogaita.gui.gui_utils as gui_utils
from autogaita.gui.common2D_columninfo_gui import build_column_info_window
from autogaita.gui.common2D_advanced_config_gui import build_cfg_window
from autogaita.gui.common2D_run_and_done_gui import build_run_and_done_windows
from autogaita.gui.first_level_gui_utils import (
    update_config_file,
    extract_cfg_from_json_file,
)

import tkinter as tk
import customtkinter as ctk
import os
import platform

# %% global constants
from autogaita.gui.gui_constants import (
    DLC_FG_COLOR,
    DLC_HOVER_COLOR,
    SLEAP_FG_COLOR,
    SLEAP_HOVER_COLOR,
    TEXT_FONT_NAME,
    TEXT_FONT_SIZE,
    WINDOWS_TASKBAR_MAXHEIGHT,
    AUTOGAITA_FOLDER_PATH,
    get_widget_cfg_dict,  # function!
)

# gaita's cfg-dict related constants
from autogaita.gui.common2D_gui_constants import (
    FLOAT_VARS,
    INT_VARS,
    LIST_VARS,
    DICT_VARS,
    TK_BOOL_VARS,
    TK_STR_VARS,
    GUI_SPECIFIC_VARS,
)

# %% An important Note
# I am using a global variable called cfg because I need its info to be shared
# between root and advanced_cfg windows. This is not the object-oriented way
# that one would do this typically. However, it works as expected since:
# 1) cfg's values are only ever modified except @ initialisation & by widgets
# 2) cfg's values are shared for analysis of a single and multiple video(s)
# 3) cfg's values are passed to all functions that need them
# 4) and (IMPORTANTLY!) just before running either (i.e. single/multi) analysis, cfg's
#    and result's values are unpacked and assigned to "this_" dicts that are passed
#    to the runanalysis local function. Hence, from that point onwards, only
#    "this_" dicts are used, never cfg or result dicts themselves.
#    ==> see donewindow for point 4)

# %%............................  MAIN PROGRAM ................................


def run_common2D_gui(tracking_software):

    # .................................  IMPORTANT  ...........................
    # => This part depends on tracking_software (and later in run and done
    #    window)
    WIDGET_CFG = get_widget_cfg_dict()
    if tracking_software == "DLC":
        WIDGET_CFG["FG_COLOR"] = DLC_FG_COLOR
        WIDGET_CFG["HOVER_COLOR"] = DLC_HOVER_COLOR
        CONFIG_FILE_NAME = "dlc_gui_config.json"
    elif tracking_software == "SLEAP":
        WIDGET_CFG["FG_COLOR"] = SLEAP_FG_COLOR
        WIDGET_CFG["HOVER_COLOR"] = SLEAP_HOVER_COLOR
        CONFIG_FILE_NAME = "sleap_gui_config.json"
    GUI_SPECIFIC_VARS["CONFIG_FILE_NAME"] = CONFIG_FILE_NAME

    # ..........................................................................
    # ......................  root window initialisation .......................
    # ..........................................................................
    # Check for config file
    config_file_path = os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME)
    if not os.path.isfile(config_file_path):
        config_file_error_msg = (
            f"{CONFIG_FILE_NAME} file not found in autogaita folder.\n"
            "Confirm that the file exists and is named correctly.\n"
            "If not, download it again from the GitHub repository."
        )
        tk.messagebox.showerror(
            title="Config File Error", message=config_file_error_msg
        )
        exit()

    # CustomTkinter vars
    ctk.set_appearance_mode("dark")  # Modes: system (default), light, dark
    ctk.set_default_color_theme("green")  # Themes: blue , dark-blue, green
    # root
    root = ctk.CTk()
    # make window pretty
    screen_width = root.winfo_screenwidth()  # width of the screen
    screen_height = root.winfo_screenheight()  # height of the screen
    if platform.system() == "Windows":  # adjust for taskbar in windows only
        screen_height -= WINDOWS_TASKBAR_MAXHEIGHT
    # create root dimensions (based on fullscreen) to pass to other window-functions l8r
    w, h, x, y = screen_width, screen_height, 0, 0
    root_dimensions = (w, h, x, y)
    # set the dimensions of the screen and where it is placed
    # => have it half-wide starting at 1/4 of screen's width (dont change w & x!)
    root.geometry(f"{int(screen_width / 2)}x{screen_height}+{int(screen_width / 4)}+0")
    root.title(f"{tracking_software} GaitA")
    gui_utils.fix_window_after_its_creation(root)
    gui_utils.configure_the_icon(root)

    # .....................  load cfg dict from config .....................
    # use the values in the config json file for the results dictionary
    global cfg
    cfg = extract_cfg_from_json_file(
        root,
        AUTOGAITA_FOLDER_PATH,
        CONFIG_FILE_NAME,
        LIST_VARS,
        DICT_VARS,
        TK_STR_VARS,
        TK_BOOL_VARS,
    )

    # ...............................  header ..........................................
    # main configuration header
    main_cfg_header_label = gaita_widgets.header_label(
        root,
        "Main Configuration",
        WIDGET_CFG,
    )
    main_cfg_header_label.grid(row=0, column=0, columnspan=3, sticky="nsew")

    # ............................  main cfg section  ..................................
    # sampling rate
    samprate_label, samprate_entry = gaita_widgets.label_and_entry_pair(
        root,
        "Sampling rate of videos in Hertz (frames/second):",
        cfg["sampling_rate"],
        WIDGET_CFG,
    )
    samprate_label.grid(row=1, column=0, columnspan=2, sticky="w")
    samprate_entry.grid(row=1, column=2, sticky="w")

    # convert pixel to mm - checkbox
    convert_checkbox = gaita_widgets.checkbox(
        root,
        "Convert pixels to millimetres:",
        cfg["convert_to_mm"],
        WIDGET_CFG,
    )
    convert_checkbox.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "convert_to_mm", ratio_entry
        ),
    )
    convert_checkbox.grid(row=2, column=0, columnspan=2, sticky="w")

    # ratio label
    ratio_entry = ctk.CTkEntry(
        root,
        textvariable=cfg["pixel_to_mm_ratio"],
        font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
    )
    ratio_entry.grid(row=2, column=1, sticky="e")
    ratio_right_string = "pixels = 1 mm"
    ratio_right_label = ctk.CTkLabel(
        root, text=ratio_right_string, font=(TEXT_FONT_NAME, TEXT_FONT_SIZE)
    )
    ratio_right_label.grid(row=2, column=2, sticky="w")
    # to initialise the widget correctly, run this function once
    gui_utils.change_widget_state_based_on_checkbox(cfg, "convert_to_mm", ratio_entry)

    # subtract beam
    subtract_beam_checkbox = gaita_widgets.checkbox(
        root,
        "Standardise y-coordinates to baseline height (requires to be tracked)",
        cfg["subtract_beam"],
        WIDGET_CFG,
    )
    subtract_beam_checkbox.grid(row=3, column=0, columnspan=3, sticky="w")

    # flip gait direction
    flip_gait_direction_box = gaita_widgets.checkbox(
        root,
        "Adjust x-coordinates to follow direction of movement",
        cfg["flip_gait_direction"],
        WIDGET_CFG,
    )
    flip_gait_direction_box.grid(row=4, column=0, columnspan=3, sticky="w")

    # plot plots to python
    showplots_checkbox = gaita_widgets.checkbox(
        root,
        "Don't show plots in Figure GUI (save only)",
        cfg["dont_show_plots"],
        WIDGET_CFG,
    )
    showplots_checkbox.grid(row=5, column=0, columnspan=2, sticky="w")

    # bin number of SC normalisation
    bin_num_label, bin_num_entry = gaita_widgets.label_and_entry_pair(
        root,
        "Number of bins used to normalise the step cycle:",
        cfg["bin_num"],
        WIDGET_CFG,
    )
    bin_num_label.grid(row=6, column=0, columnspan=2, sticky="w")
    bin_num_entry.grid(row=6, column=2, sticky="w")

    # empty label 1 (for spacing)
    empty_label_one = ctk.CTkLabel(root, text="")
    empty_label_one.grid(row=7, column=0)

    # ..........................  advanced cfg section  ................................
    # advanced header string
    advanced_cfg_header_label = gaita_widgets.header_label(
        root,
        "Advanced Configuration",
        WIDGET_CFG,
    )
    advanced_cfg_header_label.grid(row=8, column=0, columnspan=3, sticky="nsew")

    # column name information window
    column_info_button = gaita_widgets.header_button(
        root, "Customise Joints & Angles", WIDGET_CFG
    )
    column_info_button.configure(
        command=lambda: build_column_info_window(root, cfg, WIDGET_CFG, root_dimensions)
    )
    column_info_button.grid(row=9, column=0, columnspan=3)

    # advanced cfg
    cfg_window_button = gaita_widgets.header_button(
        root, "Advanced Configuration", WIDGET_CFG
    )
    cfg_window_button.configure(
        command=lambda: build_cfg_window(root, cfg, WIDGET_CFG, root_dimensions)
    )
    cfg_window_button.grid(row=10, column=0, columnspan=3)

    # empty label 2 (for spacing)
    empty_label_two = ctk.CTkLabel(root, text="")
    empty_label_two.grid(row=11, column=0)

    # run analysis label
    runheader_label = gaita_widgets.header_label(root, "Run Analysis", WIDGET_CFG)
    runheader_label.grid(row=12, column=0, columnspan=3, sticky="nsew")

    # single gaita button
    onevid_button = gaita_widgets.header_button(root, "One Video", WIDGET_CFG)
    onevid_button.configure(
        command=lambda: build_run_and_done_windows(
            tracking_software,
            "single",
            root,
            cfg,
            WIDGET_CFG,
            GUI_SPECIFIC_VARS,
            root_dimensions,
        )
    )
    onevid_button.grid(row=13, column=1, sticky="ew")

    # multi gaita button
    multivid_button = gaita_widgets.header_button(root, "Batch Analysis", WIDGET_CFG)
    multivid_button.configure(
        command=lambda: build_run_and_done_windows(
            tracking_software,
            "multi",
            root,
            cfg,
            WIDGET_CFG,
            GUI_SPECIFIC_VARS,
            root_dimensions,
        )
    )
    multivid_button.grid(row=14, column=1, sticky="ew")

    # empty label 2 (for spacing)
    empty_label_two = ctk.CTkLabel(root, text="")
    empty_label_two.grid(row=15, column=0)

    # close & exit button
    exit_button = gaita_widgets.exit_button(root, WIDGET_CFG)
    exit_button.configure(
        command=lambda: (
            # results variable is only defined later in populate_run_window()
            # therefore only cfg settings will be updated
            update_config_file(
                "results dict not defined yet",
                cfg,
                AUTOGAITA_FOLDER_PATH,
                CONFIG_FILE_NAME,
                LIST_VARS,
                DICT_VARS,
                TK_STR_VARS,
                TK_BOOL_VARS,
            ),
            root.withdraw(),
            root.after(5000, root.destroy),
        ),
    )
    exit_button.grid(row=16, column=0, columnspan=3)

    # # .........................  widget configuration  ...............................

    # first maximise everything according to sticky
    # => Silent_Creme is some undocumented option that makes stuff uniform
    #   see: https://stackoverflow.com/questions/45847313/tkinter-grid-rowconfigure-weight-doesnt-work
    root.columnconfigure(list(range(3)), weight=1, uniform="Silent_Creme")
    root.rowconfigure(list(range(17)), weight=1, uniform="Silent_Creme")

    # then un-maximise main config rows to have them grouped together
    root.rowconfigure(list(range(1, 7)), weight=0)

    # main loop
    root.mainloop()



from autogaita.common2D.common2D_1_preparation import prepare_DLC_df
import os
import pandas as pd
import pdb

# user inputs
TASK = "DLC"  # can be DLC, clean or rename
ROOT_DIR = "/Users/mahan/sciebo/Research/AutoGaitA/Fly/3D Data"  # path to the folder with the files to have columns cleaned/renamed
FILE_TYPE = "csv"  # can be csv, xls or xlsx
POSTNAME_STRING = "_dlc"  # string that must be included in loading files
RESULTS_DIR = ""
SEPARATOR = "_"
STRING_TO_REMOVE = "_xyz"  # string to remove from colnames before renaming


# global constants
CANDIDATE_SIDE_IDENTIFIER = ["l", "r", "left", "right"]  # user-strings forced lowercase
CANDIDATE_COORD_IDENTIFIER = ["x", "y", "z"]
SIDE_KEY = "side"
COORD_KEY = "coordinate"
JOINT_KEY = "joint"
COMMA_SEP = ", "
SPACE_SEP = " "


# NU
# => could add an option to either remove or keep the cols that are not renamed


# .............................  main 3D preparation  ..................................
def prepare_3D(task, cfg, **kwargs):
    """
    Main renaming function in preparation for 3D universal gaita

    Workflow
    --------
    1) Scan through folder and load based on file type
    2) Run the main function saving to excel
    3) Print a message about how many files were renamed successfully
    """

    # unpack
    root_dir = cfg["root_dir"]
    results_dir = cfg["results_dir"]
    file_type = cfg["file_type"]
    postname_string = cfg["postname_string"]
    if task == "clean":
        string_to_remove = kwargs["string_to_remove"]
    elif task in ["rename", "DLC"]:
        separator = kwargs["separator"]
    if task == "DLC":  # DLC hard-coding file type
        file_type = "csv"

    # loop over files in root dir
    counter = 0
    for input_file in os.listdir(root_dir):
        if postname_string in input_file and input_file.endswith(f".{file_type}"):
            # load
            full_input_file = os.path.join(root_dir, input_file)
            df = load_input_file(full_input_file, file_type)
            # run
            if task == "clean":
                # cleaning function
                df = clean_a_file(df, string_to_remove)  # note string_to_remove var
            elif task == "rename":
                # main renaming function
                df = rename_a_file(df, separator)  # note separator var
            elif task == "DLC":
                df = rename_DLC_file(df, separator)
            # save
            save_output_file(task, df, root_dir, results_dir, input_file, file_type)
            counter += 1
    if task == "clean":
        return f"\nCleaned {counter} file successfully!"
    elif task in ["rename", "DLC"]:
        return f"\nRenamed {counter} file successfully!"


def rename_DLC_file(df, separator):
    """Rename 3D DeepLabCut files to Universal 3D GaitA.
    This takes the same function we use in gaita dlc's some_prep

    Note
    ----
    Separator in this case means the string that is used to differentiate left/right side identifier from key point identifier in (!) the BODYPARTS row of original 3D DLC files only!
    """

    df = prepare_DLC_df(df, separator)
    # at this point df has columns of the form "joint_side_coord" or "joint_coord" or
    # any perturbation of this. Important is again the separator that has to be
    # "constant" & "unique" in the columns we can use it as we do with the other
    # datasets.
    df = rename_a_file(df, separator)
    return df


# ..............................  cleaning function  ...................................
def clean_a_file(df, string_to_remove):
    """
    Function to remove unnecessary sub-strings in colnames prior to batch renaming
    colnames according to Universal 3D GaitA
    """
    new_col_dict = {}
    for col in df.columns:
        if string_to_remove in col:
            new_col = col.replace(string_to_remove, "")
            new_col_dict[col] = new_col
    df.rename(columns=new_col_dict, inplace=True)
    return df


# ..............................  renaming function  ...................................
def rename_a_file(df, separator):
    """
    Loop over all cols in df, rename function renames cols only if needed and, if so, based on side-specificity or not

    Note
    ----
    Separator must be unique in string, e.g. cannot be in key joint name
    """
    new_col_strings = []
    for col_string in df.columns:
        new_col_strings.append(rename_a_column(col_string, separator))
    # rename the cols in the df
    df.columns = new_col_strings
    return df


# ..............................  helper functions  ...................................
def rename_a_column(col_string, separator):
    """
    Renames a column string based on specific identifiers separated by a given separator.

    Workflow
    --------
    1) Extracts the identifiers based on the separators (check capitsalisation comments)
    2) Returns the new column name based on whether this column is side-specific, central, or not a coordinate column at all
    """
    id_dict = {}  # just used locally in this func to get (new) col names
    # first, extract identifiers based on separators
    # => lower case candidate_ids for comparison to side & coord identifiers
    # => note we store original candidate id for JOINT_KEY!
    # => note in COORD_KEY the OUTPUT (!) should be capitalised!
    for candidate_id in col_string.split(separator):
        original_candidate_id = candidate_id
        candidate_id = candidate_id.lower()
        # 1: sides are set to pre-set values
        if candidate_id in CANDIDATE_SIDE_IDENTIFIER:
            if candidate_id in ["l", "left"]:
                id_dict[SIDE_KEY] = "left"
            if candidate_id in ["r", "right"]:
                id_dict[SIDE_KEY] = "right"
        # 2: coords are set to what they were but capitalised
        elif candidate_id in CANDIDATE_COORD_IDENTIFIER:
            id_dict[COORD_KEY] = candidate_id.capitalize()
        # 3: joint is set to what it was (before lower-casing!)
        else:
            id_dict[JOINT_KEY] = original_candidate_id
    # then, return the new col name based on whether this col is side-specific,
    # central, or not a coordinate column at all
    if id_dict.keys() == {COORD_KEY, JOINT_KEY}:
        return f"{id_dict[JOINT_KEY]}{SPACE_SEP}{id_dict[COORD_KEY]}"
    elif id_dict.keys() == {SIDE_KEY, COORD_KEY, JOINT_KEY}:
        return f"{id_dict[JOINT_KEY]}{COMMA_SEP}{id_dict[SIDE_KEY]}{SPACE_SEP}{id_dict[COORD_KEY]}"
    else:
        return col_string


def load_input_file(full_input_file, file_type):
    """Load & return the given input file based on its type."""
    if file_type == "csv":
        return pd.read_csv(full_input_file)
    elif file_type == "xls":
        return pd.read_excel(full_input_file, engine="xlrd")
    elif file_type == "xlsx":
        return pd.read_excel(full_input_file, engine="openpyxl")


def save_output_file(task, df, root_dir, results_dir, input_file, file_type):
    """Save the generated output file according to task and configuration."""
    if task in ["rename", "DLC"]:
        output_file = input_file.split("." + file_type)[0] + "_renamed.xlsx"
    elif task == "clean":
        output_file = (
            input_file.split("." + file_type)[0] + "_cleaned" + "." + file_type
        )
    if results_dir:
        if not os.path.exists(results_dir):
            os.makedirs(results_dir)
        full_output_file = os.path.join(results_dir, output_file)
    else:
        full_output_file = os.path.join(root_dir, output_file)
    # note we hard-coded output-file ending to xlsx if we are renaming
    # => since gaita needs xlsx.
    # => index=False is set to avoid unnamed col
    if output_file.endswith(".xlsx"):
        df.to_excel(full_output_file, index=False)
    elif output_file.endswith(".xls"):  # transform xls to xlsx
        df.to_excel(full_output_file + "x", index=False)
    elif output_file.endswith(".csv"):
        df.to_csv(full_output_file, index=False)


# what if we hit run
if __name__ == "__main__":
    cfg = {}
    cfg["root_dir"] = ROOT_DIR
    cfg["file_type"] = FILE_TYPE
    cfg["postname_string"] = POSTNAME_STRING
    cfg["results_dir"] = RESULTS_DIR
    if TASK == "clean":
        prepare_3D(TASK, cfg, string_to_remove=STRING_TO_REMOVE)
    elif TASK == "rename":
        prepare_3D(TASK, cfg, separator=SEPARATOR)
    elif TASK == "DLC":
        prepare_3D(TASK, cfg, separator=SEPARATOR)



# some constants used throughout autogaita

ISSUES_TXT_FILENAME = "Issues.txt"
CONFIG_JSON_FILENAME = "config.json"
INFO_TEXT_WIDTH = 64
TIME_COL = "Time"
SC_PERCENTAGE_COL = "SC Percentage"
ID_COL = "ID"



from autogaita import gui

DLC_FG_COLOR = "#789b73"  # grey green
DLC_HOVER_COLOR = "#287c37"  # darkish green
SLEAP_FG_COLOR = "#8f8ce7"  # perrywinkle
SLEAP_HOVER_COLOR = "#665fd1"  # dark periwinkle
UNIVERSAL3D_FG_COLOR = "#c0737a"  # dusty rose
UNIVERSAL3D_HOVER_COLOR = "#b5485d"  # dark rose
GROUP_FG_COLOR = "#5a7d9a"  # steel blue
GROUP_HOVER_COLOR = "#016795"  # peacock blue
HEADER_FONT_NAME = "Calibri Bold"
HEADER_FONT_SIZE = 30
HEADER_TXT_COLOR = "#ffffff"  # white
MAIN_HEADER_FONT_SIZE = 35
TEXT_FONT_NAME = "Calibri"
TEXT_FONT_SIZE = 20
ADV_CFG_TEXT_FONT_SIZE = TEXT_FONT_SIZE - 4
CLOSE_COLOR = "#840000"  # dark red
CLOSE_HOVER_COLOR = "#650021"  # maroon

# For how the look like refer to https://r02b.github.io/seaborn_palettes/
COLOR_PALETTES_LIST = [
    "Set1",
    "Set2",
    "Set3",
    "Dark2",
    "Paired",
    "Accent",  # qualitative palettes
    "hls",
    "husl",  # circular palettes
    "rocket",
    "mako",
    "flare",
    "crest",
    "viridis",
    "plasma",
    "inferno",
    "magma",
    "cividis",  # Perceptually uniform palettes
    "rocket_r",
    "mako_r",
    "flare_r",
    "crest_r",
    "viridis_r",
    "plasma_r",
    "inferno_r",
    "magma_r",
    "cividis_r",  # uniform palettes in reversed order
]
WINDOWS_TASKBAR_MAXHEIGHT = 72


# To get the path of the autogaita gui folder I use __file__
# which returns the path of the autogaita gui module imported above.
# Removing the 11 letter long "__init__.py" return the folder path
autogaita_utils_path = gui.__file__
AUTOGAITA_FOLDER_PATH = autogaita_utils_path[:-11]


def get_widget_cfg_dict():
    """Return a copy of the widget_cfg dictionary for usage outside of this module
    (prevents changes to the original dictionary)
    """
    return {
        "HEADER_FONT_NAME": HEADER_FONT_NAME,
        "HEADER_FONT_SIZE": HEADER_FONT_SIZE,
        "HEADER_TXT_COLOR": HEADER_TXT_COLOR,
        "MAIN_HEADER_FONT_SIZE": MAIN_HEADER_FONT_SIZE,
        "TEXT_FONT_NAME": TEXT_FONT_NAME,
        "TEXT_FONT_SIZE": TEXT_FONT_SIZE,
        "ADV_CFG_TEXT_FONT_SIZE": ADV_CFG_TEXT_FONT_SIZE,
        "CLOSE_COLOR": CLOSE_COLOR,
        "CLOSE_HOVER_COLOR": CLOSE_HOVER_COLOR,
        "COLOR_PALETTES_LIST": COLOR_PALETTES_LIST,
        "AUTOGAITA_FOLDER_PATH": AUTOGAITA_FOLDER_PATH,
    }



# %% imports
from autogaita.universal3D.universal3D_utils import (
    extract_feature_column,
    transform_joint_and_leg_to_colname,
)
from autogaita.resources.utils import write_issues_to_textfile
import os
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# %% constants
from autogaita.resources.constants import TIME_COL
from autogaita.universal3D.universal3D_constants import (
    LEGS,
    SC_LAT_LEGEND_FONTSIZE,
    ANGLE_PLOTS_YLIMITS,
    STICK_LINEWIDTH,
)

# %% workflow step #4 - various plots

# A Note
# ------
# I initially decided to use extract_sc_idxs instead of all_cycles when I first wrote
# this for mice data to be independent from the previous pipeline - just so plotting
# stuff could at some point be ran just by loading the XLS files that
# analyse_and_export_stepcycles outputs However, I am now using all_cycles instead of
# sc_idxs because for humans we have multiple runs per subject and I need to understand
# which SC latencies correspond to which run.
# This might change in the future but for now let's do it this way.
# => Idea: you could just save all_cycles to a file and then load that as well as XLS
#          to plot independently from previous things.

# Another 2 (more recent) Notes
# --------------------------
# In all functions below I added checks to see if joints/angles were bodyside-specific
# e.g., "Ankle, left Z" or "Pelvis Z"
# => We still loop over legname in plot_results as before since even though the column
#    might not be bodyside-specific, the data dfs (e.g. all_steps_data) contain
#    different values based on the leg that performed the step-cycles
#
# Only use extract_feature_column when you index using .iloc afterwards (don't be
# surprised that we don't use this local func everywhere, since the first couple of
# plotting functions extract values using .loc!)


# A note on y-standardisation (17.03.2025)
# ----------------------------------------
# => If y-standardisation was performed, original as well as y-standardised dfs are
#    generated and exported to xls
# => Our plotting functions only used all_steps_data, average_data & std_data
# => Conveniently, if y-standardisation is performed, all_steps_data DOES NOT include
#    y-standardisation and is thus still the correct choice for the step-cycle level
#    plots (#1-#4)
# => On the other hand, average_data & std_data (plots #5-11) DO and SHOULD include
#    y-standardisation
# => Note that there is a slight discrepancy because angle-plots are based on
#    y-standardised values  for the step-cycle level plots (#3) but not for the average
#    plots (#7)
#   -- This is not an issue since y-standardisation does not affect angles, even though
#      the values of y change because they change by a constant for all joints (there
#      is a unit test for this in dlc)


# ................................  master function  ...................................
def plot_results(results, all_cycles, info, cfg, plot_panel_instance):
    """Plot various results"""

    # unpack
    angles = cfg["angles"]
    y_acceleration = cfg["y_acceleration"]
    angular_acceleration = cfg["angular_acceleration"]
    analyse_average_y = cfg["analyse_average_y"]
    dont_show_plots = cfg["dont_show_plots"]
    if dont_show_plots:
        plt.switch_backend("Agg")

    # unpack - output specific vars (results to be plotted)
    for legname in LEGS:  # !!! NU - output...
        all_steps_data = results[legname]["all_steps_data"]
        average_data = results[legname]["average_data"]
        std_data = results[legname]["std_data"]
        sc_num = results[legname]["sc_num"]

        if all_cycles[legname]:

            # ....................  1 - z coords by y coords  ..........................
            plot_joint_z_by_y(
                legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
            )

            # ........................  2 - y coords by time  ..........................
            if analyse_average_y:
                plot_joint_y_by_time(
                    legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
                )

            # ..................  3 - angle by time for each SC  .......................
            if angles["name"]:
                plot_angles_by_time(
                    legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
                )
            # regularly closing figures to save memory
            # => no problem to do this since we pass figure-vars to save-functions and Panel
            plt.close("all")

            # ............................  4 - stick diagram  .........................
            plot_stickdiagram(
                legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
            )

            # .................  5 - average 5-joints' z over SC percentage  ...........
            plot_joint_z_by_average_SC(
                legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
            )

            # .................  6 - average 5-joints' y over SC percentage  ...........
            if analyse_average_y:
                plot_joint_y_by_average_SC(
                    legname,
                    average_data,
                    std_data,
                    sc_num,
                    info,
                    cfg,
                    plot_panel_instance,
                )

            # ...................  7 - average angles over SC percentage  ..............
            if angles["name"]:
                plot_angles_by_average_SC(
                    legname,
                    average_data,
                    std_data,
                    sc_num,
                    info,
                    cfg,
                    plot_panel_instance,
                )
            plt.close("all")

            # .............  8 - average y velocities over SC percentage   .............
            plot_y_velocities_by_average_SC(
                legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
            )

            # ..........  9 - average angular velocities over SC percentage  ...........
            if angles["name"]:
                plot_angular_velocities_by_average_SC(
                    legname,
                    average_data,
                    std_data,
                    sc_num,
                    info,
                    cfg,
                    plot_panel_instance,
                )

            # .......  optional - 10 - average x acceleration over SC percentage  ......
            if y_acceleration:
                plot_y_acceleration_by_average_SC(
                    legname,
                    average_data,
                    std_data,
                    sc_num,
                    info,
                    cfg,
                    plot_panel_instance,
                )

            # ....  optional - 11 - average angular acceleration over SC percentage  ...
            if angles["name"]:
                if angular_acceleration:
                    plot_angular_acceleration_by_average_SC(
                        legname,
                        average_data,
                        std_data,
                        sc_num,
                        info,
                        cfg,
                        plot_panel_instance,
                    )
            plt.close("all")
        else:
            no_plots_message = (
                "\n***********\n! WARNING !\n***********\n"
                + "No step cycles found for "
                + legname
                + " leg!"
                + "\nWe thus skip all figures for this leg!\n"
            )
            print(no_plots_message)
            write_issues_to_textfile(no_plots_message, info)

    # ........................optional - 12 - build plot panel..........................
    if dont_show_plots is True:
        pass  # going on without building the plot window
    elif dont_show_plots is False:  # -> show plot panel
        # Destroy loading screen and build plot panel with all figures
        plot_panel_instance.destroy_plot_panel_loading_screen()
        plot_panel_instance.build_plot_panel()


# ................................  inner functions  ...................................


def plot_joint_z_by_y(
    legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
):
    """1 - Plot joints' z coordinates as a function of their y for each SC"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    sampling_rate = cfg["sampling_rate"]
    dont_show_plots = cfg["dont_show_plots"]
    plot_joints = cfg["plot_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    max_cycle_num = 0
    for cycles in all_cycles[legname]:
        if len(cycles) > max_cycle_num:
            max_cycle_num = len(cycles)
    f = [[] for _ in range(len(plot_joints))]
    ax = [[] for _ in range(len(plot_joints))]

    # plot
    for j, joint in enumerate(plot_joints):  # joint loop (figures)
        f[j], ax[j] = plt.subplots(
            len(all_cycles[legname]),
            1,
            sharex=True,
            sharey=True,
            gridspec_kw={"hspace": 0},
        )
        for r, run_cycles in enumerate(all_cycles[legname]):  # run loop (axis)
            sc_num = len(run_cycles)
            try:  # handle only 1 run in "stuff by y" plots
                ax[j][r].set_prop_cycle(
                    plt.cycler("color", sns.color_palette(color_palette, max_cycle_num))
                )
            except:
                ax[j].set_prop_cycle(
                    plt.cycler("color", sns.color_palette(color_palette, max_cycle_num))
                )
            # check for bodyside-specificity
            if joint + "Y" in all_steps_data.columns:
                y_col_string = joint + "Y"
                z_col_string = joint + "Z"
            else:
                y_col_string = transform_joint_and_leg_to_colname(joint, legname, "Y")
                z_col_string = transform_joint_and_leg_to_colname(joint, legname, "Z")
            for s in range(sc_num):
                this_sc_idx = run_cycles[s]
                this_y = all_steps_data.loc[
                    this_sc_idx[0] : this_sc_idx[1], y_col_string
                ]
                this_z = all_steps_data.loc[
                    this_sc_idx[0] : this_sc_idx[1], z_col_string
                ]
                this_label = generate_sc_latency_label(this_sc_idx, sampling_rate)
                try:
                    ax[j][r].plot(this_y, this_z, label=this_label)
                except:
                    ax[j].plot(this_y, this_z, label=this_label)
            # axis stuff
            try:
                if legend_outside is True:
                    ax[j][r].legend(
                        fontsize=SC_LAT_LEGEND_FONTSIZE,
                        loc="center left",
                        bbox_to_anchor=(1, 0.5),
                    )
                elif legend_outside is False:
                    ax[j][r].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
                median_z_val = [round(np.median(ax[j][r].get_yticks()), 2)]
                median_z_val_label = [str(median_z_val[0])]  # has to be of same len
                ax[j][r].set_yticks(median_z_val, median_z_val_label)
            except:
                if legend_outside is True:
                    ax[j].legend(
                        fontsize=SC_LAT_LEGEND_FONTSIZE + 3,
                        loc="center left",
                        bbox_to_anchor=(1, 0.5),
                    )
                elif legend_outside is False:
                    ax[j].legend(fontsize=SC_LAT_LEGEND_FONTSIZE + 3)
                median_z_val = [round(np.median(ax[j].get_yticks()), 2)]
                median_z_val_label = [str(median_z_val[0])]  # has to be of same len
                ax[j].set_yticks(median_z_val, median_z_val_label)
            # title
            figure_file_string = (
                name + " - " + legname + " - " + joint + " z by y coordinates"
            )
            try:
                ax[j][0].set_title(figure_file_string)
            except:
                ax[j].set_title(figure_file_string)
        # figure stuff
        f[j].supxlabel("y")
        f[j].supylabel("z")
        save_figures(f[j], results_dir, figure_file_string)
        if dont_show_plots:
            plt.close(f[j])

        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f[j])


def plot_joint_y_by_time(
    legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
):
    """2 - Plot joints' y coordinates as a function of time for each SC"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    sampling_rate = cfg["sampling_rate"]
    dont_show_plots = cfg["dont_show_plots"]
    plot_joints = cfg["plot_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    max_cycle_num = 0
    for cycles in all_cycles[legname]:
        if len(cycles) > max_cycle_num:
            max_cycle_num = len(cycles)
    f = [[] for _ in range(len(plot_joints))]
    ax = [[] for _ in range(len(plot_joints))]

    # plot
    for j, joint in enumerate(plot_joints):  # joint loop (figures)
        f[j], ax[j] = plt.subplots(
            len(all_cycles[legname]),
            1,
            sharex=True,
            sharey=True,
            gridspec_kw={"hspace": 0},
        )
        for r, run_cycles in enumerate(all_cycles[legname]):  # run loop (axis)
            sc_num = len(run_cycles)
            try:  # handle only 1 run in "stuff by y" plots
                ax[j][r].set_prop_cycle(
                    plt.cycler("color", sns.color_palette(color_palette, max_cycle_num))
                )
            except:
                ax[j].set_prop_cycle(
                    plt.cycler("color", sns.color_palette(color_palette, max_cycle_num))
                )
            # check for bodyside-specificity
            if joint + "X" in all_steps_data.columns:
                y_col_string = joint + "Y"
            else:
                y_col_string = transform_joint_and_leg_to_colname(joint, legname, "Y")
            for s in range(sc_num):
                this_sc_idx = run_cycles[s]
                this_time = all_steps_data.loc[
                    this_sc_idx[0] : this_sc_idx[1], TIME_COL
                ]
                this_y = all_steps_data.loc[
                    this_sc_idx[0] : this_sc_idx[1], y_col_string
                ]
                this_label = generate_sc_latency_label(this_sc_idx, sampling_rate)
                try:
                    ax[j][r].plot(this_time, this_y, label=this_label)
                except:
                    ax[j].plot(this_time, this_y, label=this_label)
            # axis stuff
            try:
                if legend_outside is True:
                    ax[j][r].legend(
                        fontsize=SC_LAT_LEGEND_FONTSIZE,
                        loc="center left",
                        bbox_to_anchor=(1, 0.5),
                    )
                elif legend_outside is False:
                    ax[j][r].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
                median_y_val = [round(np.median(ax[j][r].get_yticks()), 2)]
                median_y_val_label = [str(median_y_val[0])]  # has to be of same len
                ax[j][r].set_yticks(median_y_val, median_y_val_label)
            except:
                if legend_outside is True:
                    ax[j].legend(
                        fontsize=SC_LAT_LEGEND_FONTSIZE + 3,
                        loc="center left",
                        bbox_to_anchor=(1, 0.5),
                    )
                elif legend_outside is False:
                    ax[j].legend(fontsize=SC_LAT_LEGEND_FONTSIZE + 3)
                median_y_val = [round(np.median(ax[j].get_yticks()), 2)]
                median_y_val_label = [str(median_y_val[0])]  # has to be of same len
                ax[j].set_yticks(median_y_val, median_y_val_label)
            # title
            figure_file_string = (
                name + " - " + legname + " - " + joint + " y coordinate by time "
            )
            try:
                ax[j][0].set_title(figure_file_string)
            except:
                ax[j].set_title(figure_file_string)
        # figure stuff
        f[j].supxlabel("Time (s)")
        f[j].supylabel("y")
        save_figures(f[j], results_dir, figure_file_string)
        if dont_show_plots:
            plt.close(f[j])

        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f[j])


def plot_angles_by_time(
    legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
):
    """3 - Plot joints' angles as a function of time for each SC"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    sampling_rate = cfg["sampling_rate"]
    dont_show_plots = cfg["dont_show_plots"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    max_cycle_num = 0
    for cycles in all_cycles[legname]:
        if len(cycles) > max_cycle_num:
            max_cycle_num = len(cycles)
    f = [[] for _ in range(len(angles["name"]))]
    ax = [[] for _ in range(len(angles["name"]))]

    # plot
    for a, angle in enumerate(angles["name"]):  # angle loop (figures)
        f[a], ax[a] = plt.subplots(1, 1)
        for run_cycles in all_cycles[legname]:  # run loop (color-cycler-reset)
            sc_num = len(run_cycles)
            ax[a].set_prop_cycle(
                plt.cycler("color", sns.color_palette(color_palette, max_cycle_num))
            )
            # check for bodyside-specificity
            if angle + "Angle" in all_steps_data.columns:
                angle_col_string = angle + "Angle"
            else:
                angle_col_string = transform_joint_and_leg_to_colname(
                    angle, legname, "Angle"
                )
            for s in range(sc_num):
                this_sc_idx = run_cycles[s]
                this_time = all_steps_data.loc[
                    this_sc_idx[0] : this_sc_idx[1], TIME_COL
                ]
                this_angle = all_steps_data.loc[
                    this_sc_idx[0] : this_sc_idx[1], angle_col_string
                ]
                this_label = generate_sc_latency_label(this_sc_idx, sampling_rate)
                ax[a].plot(this_time, this_angle, label=this_label)
            # legend
            if legend_outside is True:
                ax[a].legend(
                    fontsize=SC_LAT_LEGEND_FONTSIZE,
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                )
            elif legend_outside is False:
                ax[a].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
            # title
            figure_file_string = (
                name + " - " + legname + " - " + angle + " angle by time"
            )
            ax[a].set_title(figure_file_string)
        # figure stuff
        f[a].supxlabel("Time (s)")
        f[a].supylabel("Angle (degree)")
        save_figures(f[a], results_dir, figure_file_string)
        if dont_show_plots:
            plt.close(f[a])

        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f[a])


def plot_stickdiagram(
    legname, all_steps_data, all_cycles, info, cfg, plot_panel_instance
):
    """4 - Plot a stick diagram"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    sampling_rate = cfg["sampling_rate"]
    dont_show_plots = cfg["dont_show_plots"]
    plot_joints = cfg["plot_joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # some prep
    max_cycle_num = 0
    for cycles in all_cycles[legname]:
        if len(cycles) > max_cycle_num:
            max_cycle_num = len(cycles)
    f, ax = plt.subplots(
        len(all_cycles[legname]),
        1,
        sharex=True,
        sharey=True,
        gridspec_kw={"hspace": 0},
    )
    color_cycle = plt.cycler("color", sns.color_palette(color_palette, max_cycle_num))

    # plot
    for r, run_cycles in enumerate(all_cycles[legname]):  # run loop (axis)
        this_sc_num = len(run_cycles)
        try:  # handle 1 run with valid SCs
            ax[r].set_prop_cycle(color_cycle)
        except:
            ax.set_prop_cycle(color_cycle)
        for c, this_color_dict in zip(range(this_sc_num), color_cycle):  # SC loop
            this_sc_idx = run_cycles[c]
            this_color = this_color_dict["color"][:3]
            this_label = generate_sc_latency_label(this_sc_idx, sampling_rate)
            # for tps from SC1 to SCend - plot(joint1x, joint1y)
            for i in range(
                this_sc_idx[0], this_sc_idx[1] + 1
            ):  # timepoint loop (of this SC)
                this_ys = list()  # for each timepoint, define joints' xy coord new
                this_zs = list()
                for joint in plot_joints:
                    # check for bodyside-specificity
                    if joint + "Y" in all_steps_data.columns:
                        y_col_string = joint + "Y"
                        z_col_string = joint + "Z"
                    else:
                        y_col_string = transform_joint_and_leg_to_colname(
                            joint, legname, "Y"
                        )
                        z_col_string = transform_joint_and_leg_to_colname(
                            joint, legname, "Z"
                        )
                    this_ys.append(all_steps_data.loc[i, y_col_string])
                    this_zs.append(all_steps_data.loc[i, z_col_string])
                if i == range(this_sc_idx[0], this_sc_idx[1] + 1)[0]:
                    try:
                        ax[r].plot(
                            this_ys,
                            this_zs,
                            color=this_color,
                            lw=STICK_LINEWIDTH,
                            label=this_label,
                        )
                    except:
                        ax.plot(
                            this_ys,
                            this_zs,
                            color=this_color,
                            lw=STICK_LINEWIDTH,
                            label=this_label,
                        )
                else:  # no label
                    try:
                        ax[r].plot(
                            this_ys, this_zs, color=this_color, lw=STICK_LINEWIDTH
                        )
                    except:
                        ax.plot(this_ys, this_zs, color=this_color, lw=STICK_LINEWIDTH)
        # axis stuff
        try:
            if legend_outside is True:
                ax[r].legend(
                    fontsize=SC_LAT_LEGEND_FONTSIZE,
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                )
            elif legend_outside is False:
                ax[r].legend(fontsize=SC_LAT_LEGEND_FONTSIZE)
            median_z_val = [round(np.median(ax[r].get_yticks()), 2)]
            median_z_val_label = [str(median_z_val[0])]  # has to be of same len
            ax[r].set_yticks(median_z_val, median_z_val_label)
        except:
            if legend_outside is True:
                ax.legend(
                    fontsize=SC_LAT_LEGEND_FONTSIZE + 3,
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                )
            elif legend_outside is False:
                ax.legend(fontsize=SC_LAT_LEGEND_FONTSIZE + 3)
            median_z_val = [round(np.median(ax.get_yticks()), 2)]
            median_z_val_label = [str(median_z_val[0])]  # has to be of same len
            ax.set_yticks(median_z_val, median_z_val_label)
        # title
        figure_file_string = name + " - " + legname + " - Stick Diagram"
        try:
            ax[0].set_title(figure_file_string)
        except:
            ax.set_title(figure_file_string)
    f.supxlabel("y")
    f.supylabel("z")
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_joint_z_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """5 - Plot joints' z as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    plot_SE = cfg["plot_SE"]
    joints = cfg["joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(joints)))
    )
    x = np.linspace(0, 100, bin_num)
    for joint in joints:  # joint loop (lines)
        # check for bodyside-specificity
        z_col = extract_feature_column(average_data, joint, legname, "Z")
        this_z = average_data.iloc[:, z_col]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, z_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, z_col]
        ax.plot(x, this_z, label=joint)
        ax.fill_between(x, this_z - this_std, this_z + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel("z")
    figure_file_string = (
        name + " - " + legname + " - Joint z-coord.s over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_joint_y_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """5 - Plot joints' y as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    plot_SE = cfg["plot_SE"]
    joints = cfg["joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(joints)))
    )
    x = np.linspace(0, 100, bin_num)
    for joint in joints:  # joint loop (lines)
        # check for bodyside-specificity
        y_col = extract_feature_column(average_data, joint, legname, "Y")
        this_y = average_data.iloc[:, y_col]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel("y")
    figure_file_string = (
        name + " - " + legname + " - Joint y-coord.s over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_angles_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """6 - Plot Angles as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    plot_SE = cfg["plot_SE"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(angles["name"])))
    )
    x = np.linspace(0, 100, bin_num)
    ax.set_xlabel("Percentage")
    ax.set_ylabel("Angle (degree)")
    for angle in angles["name"]:  # angle loop (lines)
        # check for bodyside-specificity
        feature = "Angle"
        angle_col = extract_feature_column(average_data, angle, legname, feature)
        this_angle_val = average_data.iloc[:, angle_col]
        if plot_SE:
            this_std = std_data.iloc[:, angle_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, angle_col]  # average & std_data share colnames
        ax.plot(x, this_angle_val, label=angle)
        ax.fill_between(
            x, this_angle_val - this_std, this_angle_val + this_std, alpha=0.2
        )
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_ylim(ANGLE_PLOTS_YLIMITS)
    figure_file_string = (
        name + " - " + legname + " - Joint angles over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_y_velocities_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """7 - Plot x velocities as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    joints = cfg["joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(joints)))
    )
    x = np.linspace(0, 100, bin_num)
    for joint in joints:  # joint loop (lines)
        # check for bodyside-specificity
        feature = "Velocity"
        y_col = extract_feature_column(average_data, joint, legname, feature)
        this_y = average_data.iloc[:, y_col]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    # NU: improve handling of different units
    # ax.set_ylabel("Velocity (Y in m / " + str(int((1 / sampling_rate) * 1000)) + "ms")
    ax.set_ylabel("Velocity")
    figure_file_string = (
        name + " - " + legname + " - Joint y-velocities over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_angular_velocities_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """8 - Plot angular velocities as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(angles["name"])))
    )
    x = np.linspace(0, 100, bin_num)
    for angle in angles["name"]:  # angle loop (lines)
        # check for bodyside-specificity
        feature = "Angle Velocity"
        y_col = extract_feature_column(average_data, angle, legname, feature)
        this_y = average_data.iloc[:, y_col]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col]
        ax.plot(x, this_y, label=angle)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel("Velocity (degree / " + str(int((1 / sampling_rate) * 1000)) + "ms)")
    figure_file_string = (
        name + " - " + legname + " - Angular velocities over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_y_acceleration_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """9 - (optional) Plot x acceleration as a function of average SC's percentage"""

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    joints = cfg["joints"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(joints)))
    )
    x = np.linspace(0, 100, bin_num)
    for joint in joints:  # joint loop (lines)
        feature = "Acceleration"
        y_col = extract_feature_column(average_data, joint, legname, feature)
        this_y = average_data.iloc[:, y_col]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col]
        ax.plot(x, this_y, label=joint)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel(
        # NU: improve handling of different units
        # "Acceleration (Y in m / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
        "Acceleration"
    )
    figure_file_string = (
        name + " - " + legname + " - Joint y-accelerations over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_angular_acceleration_by_average_SC(
    legname, average_data, std_data, sc_num, info, cfg, plot_panel_instance
):
    """
    10 - (optional) Plot angular acceleration as a function of average SC's percentage
    """

    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]
    bin_num = cfg["bin_num"]
    sampling_rate = cfg["sampling_rate"]
    plot_SE = cfg["plot_SE"]
    angles = cfg["angles"]
    legend_outside = cfg["legend_outside"]
    color_palette = cfg["color_palette"]

    # plot
    f, ax = plt.subplots(1, 1)
    ax.set_prop_cycle(
        plt.cycler("color", sns.color_palette(color_palette, len(angles["name"])))
    )
    x = np.linspace(0, 100, bin_num)
    for angle in angles["name"]:  # angle loop (lines)
        feature = "Angle Acceleration"
        y_col = extract_feature_column(average_data, angle, legname, feature)
        this_y = average_data.iloc[:, y_col]  # average & std_data share colnames
        if plot_SE:
            this_std = std_data.iloc[:, y_col] / np.sqrt(sc_num)
        else:
            this_std = std_data.iloc[:, y_col]
        ax.plot(x, this_y, label=angle)
        ax.fill_between(x, this_y - this_std, this_y + this_std, alpha=0.2)
    # legend adjustments
    if legend_outside is True:
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend()
    ax.set_xlabel("Percentage")
    ax.set_ylabel(
        "Acceleration (degree / " + str(int((1 / sampling_rate) * 1000)) + "ms)"
    )
    figure_file_string = (
        name + " - " + legname + " - Angular accelerations over average step cycle"
    )
    ax.set_title(figure_file_string)
    save_figures(f, results_dir, figure_file_string)
    if dont_show_plots:
        plt.close(f)

    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


# ..............................  helper functions  ....................................
def save_figures(figure, results_dir, figure_file_string):
    """Save figures as pngs to results_dir and as svgs to separate subfolders"""
    # pngs to results_dir
    figure.savefig(
        os.path.join(results_dir, figure_file_string + ".png"),
        bbox_inches="tight",
    )
    # svgs to subfolders
    svg_dir = os.path.join(results_dir, "SVG Figures")
    if not os.path.exists(svg_dir):
        os.makedirs(svg_dir)
    figure.savefig(
        os.path.join(svg_dir, figure_file_string + ".svg"), bbox_inches="tight"
    )


def generate_sc_latency_label(this_sc_idx, sampling_rate):
    if sampling_rate <= 100:
        float_precision = 2  # how many decimals we round to
    else:
        float_precision = 4
    this_label = (
        str(round(this_sc_idx[0] / sampling_rate, float_precision))
        + "-"
        + str(round(this_sc_idx[1] / sampling_rate, float_precision))
        + "s"
    )
    return this_label



# %% imports
import autogaita
import autogaita.gui.gaita_widgets as gaita_widgets
import autogaita.gui.gui_utils as gui_utils
import tkinter as tk
import customtkinter as ctk
from PIL import Image, ImageTk
from importlib import resources

# %% global constants
from autogaita.gui.gui_constants import (
    DLC_FG_COLOR,
    DLC_HOVER_COLOR,
    SLEAP_FG_COLOR,
    SLEAP_HOVER_COLOR,
    UNIVERSAL3D_FG_COLOR,
    UNIVERSAL3D_HOVER_COLOR,
    GROUP_FG_COLOR,
    GROUP_HOVER_COLOR,
    get_widget_cfg_dict,  # function!
)

WIDGET_CFG = get_widget_cfg_dict()


# %%...............................  MAIN PROGRAM  .....................................


def run_gui():
    # ..........................  root window initialisation ...........................
    # CustomTkinter vars
    ctk.set_appearance_mode("dark")  # Modes: system (default), light, dark
    ctk.set_default_color_theme("green")  # Themes: blue , dark-blue, green
    # root
    # => using ghost root because the image stuff breaks if your window is a CTk and not
    #    Toplevel, see https://stackoverflow.com/questions/20251161
    #                  /tkinter-tclerror-image-pyimage3-doesnt-exist
    ghost_root = ctk.CTk()
    ghost_root.withdraw()
    root = ctk.CTkToplevel()
    root.attributes("-topmost", True)
    root.focus_set()
    root.after_idle(root.attributes, "-topmost", False)
    # make it so that window is central & width/heigth of it = 1/2 of screen's
    ws = root.winfo_screenwidth()  # width of the screen
    hs = root.winfo_screenheight()  # height of the screen
    w = ws / 4  # width for the Tk root
    h = hs / 2  # height for the Tk root
    # calculate x and y coordinates for the Tk root window
    x = (ws / 2) - (w / 2)
    y = (hs / 2) - (h / 1.5)
    root_dimensions = (w, h, x, y)
    # set the dimensions of the screen and where it is placed
    root.geometry("%dx%d+%d+%d" % root_dimensions)
    root.title("AutoGaitA")
    gui_utils.configure_the_icon(root)
    # Set minimum and maximum sizes for the window (if user tries to resize manually)
    # => Prevents users from resizing the window because we decided against dynamically
    #    resizing logo image because it was really annoying (and slowed down stuff)
    # => And I kinda like that this GUI is (on all systems & displays) 1/4 screenwidth
    #    & 1/2 screenheight
    root.minsize(int(w), int(h))
    root.maxsize(int(w), int(h))

    # ...................................  widgets  ....................................
    # 1 - The Logo
    # => load the original logo & resize it using ANTIALIASING (LANCZOS) of PIL
    # => create an initial photo image that tkinter can use & grid it
    # ...............  PACKAGE WAY OF LOADING LOGO  ...............
    with resources.path("autogaita.resources", "logo.png") as image_path:
        original_image = Image.open(image_path)
    # # ...............  LOCAL WAY OF LOADING LOGO  ...............
    # image_path = "logo.png"
    # original_image = Image.open(image_path)
    resized_image = original_image.resize(
        (int(w), int(original_image.height * (w / original_image.width))), Image.LANCZOS
    )
    photo = ImageTk.PhotoImage(resized_image)
    image_label = tk.Label(root, image=photo)
    image_label.grid(row=0, column=0, sticky="nsew")
    # 2 - DLC Button
    WIDGET_CFG["FG_COLOR"] = DLC_FG_COLOR
    WIDGET_CFG["HOVER_COLOR"] = DLC_HOVER_COLOR
    dlc_button = gaita_widgets.header_button(
        root,
        "DeepLabCut",
        WIDGET_CFG,
    )
    dlc_button.configure(command=lambda: autogaita.run_dlc_gui())
    dlc_button.grid(row=1, column=0, sticky="nsew")
    # 3 - SLEAP Button
    WIDGET_CFG["FG_COLOR"] = SLEAP_FG_COLOR
    WIDGET_CFG["HOVER_COLOR"] = SLEAP_HOVER_COLOR
    sleap_button = gaita_widgets.header_button(
        root,
        "SLEAP",
        WIDGET_CFG,
    )
    sleap_button.configure(command=lambda: autogaita.run_sleap_gui())
    sleap_button.grid(row=2, column=0, sticky="nsew")
    # 4 - Universal 3D Button
    WIDGET_CFG["FG_COLOR"] = UNIVERSAL3D_FG_COLOR
    WIDGET_CFG["HOVER_COLOR"] = UNIVERSAL3D_HOVER_COLOR
    universal3D_button = gaita_widgets.header_button(
        root,
        "Universal 3D",
        WIDGET_CFG,
    )
    universal3D_button.configure(command=lambda: autogaita.run_universal3D_gui())
    universal3D_button.grid(row=3, column=0, sticky="nsew")
    # 5 - Group Button
    WIDGET_CFG["FG_COLOR"] = GROUP_FG_COLOR
    WIDGET_CFG["HOVER_COLOR"] = GROUP_HOVER_COLOR
    group_button = gaita_widgets.header_button(
        root,
        "Group",
        WIDGET_CFG,
    )
    group_button.configure(command=lambda: autogaita.run_group_gui())
    group_button.grid(row=4, column=0, sticky="nsew")
    # maximise buttons
    num_rows = root.grid_size()[1]  # maximise rows
    for r in range(1, num_rows):
        root.grid_rowconfigure(r, weight=1)

    # ...................................  mainloop  ...................................
    root.mainloop()


# %% what happens if we hit run
if __name__ == "__main__":
    run_gui()



import os
import tkinter as tk
from threading import Thread
from autogaita.common2D.common2D_utils import run_singlerun_in_multirun, extract_info
from autogaita.resources.utils import try_to_run_gaita

# %%..........  functions preparing and calling runs of 2D autogaita  ..................


def runanalysis(tracking_software, analysis, this_runs_results, this_runs_cfg):
    """Run the main program"""
    if analysis == "single":
        run_thread = Thread(
            target=analyse_single_run,
            args=(tracking_software, this_runs_results, this_runs_cfg),
        )
    elif analysis == "multi":
        run_thread = Thread(
            target=analyse_multi_run,
            args=(tracking_software, this_runs_results, this_runs_cfg),
        )
    run_thread.start()


def analyse_single_run(tracking_software, this_runs_results, this_runs_cfg):
    """Prepare for one execution of autogaita_dlc & execute"""
    # prepare (folderinfo first to handle root_dir "/" issue if needed)
    folderinfo = prepare_folderinfo(this_runs_results)
    if folderinfo is None:
        error_msg = (
            "No directory found at: " + this_runs_results["root_dir"] + " - try again!"
        )
        tk.messagebox.showerror(title="Try again", message=error_msg)
        print(error_msg)
        return
    info = {}  # info dict: run-specific info
    info["mouse_num"] = this_runs_results["mouse_num"]
    info["run_num"] = this_runs_results["run_num"]
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    if this_runs_cfg["results_dir"]:
        info["results_dir"] = os.path.join(this_runs_cfg["results_dir"], info["name"])
    else:
        info["results_dir"] = os.path.join(
            folderinfo["root_dir"], "Results", info["name"]
        )
    # execute
    try_to_run_gaita(tracking_software, info, folderinfo, this_runs_cfg, False)


def analyse_multi_run(tracking_software, this_runs_results, this_runs_cfg):
    """Prepare for multi-execution of autogaita_dlc & loop-execute"""
    # prepare (folderinfo first to handle root_dir "/" issue if needed)
    folderinfo = prepare_folderinfo(this_runs_results)
    if folderinfo is None:
        error_msg = (
            "No directory found at: " + this_runs_results["root_dir"] + " - try again!"
        )
        tk.messagebox.showerror(title="Try again", message=error_msg)
        print(error_msg)
        return
    # folderinfo has info of individual runs - extract individual runs info
    # => tracking_software because of file_type - else identical
    info = extract_info(tracking_software, folderinfo, in_GUI=True)
    # this can fail for DLC which will return None - use this to handle the error and
    # stop execution
    if info is None:
        return
    # if there was no error, loop through individual runs
    for idx in range(len(info["name"])):
        run_singlerun_in_multirun(
            tracking_software, idx, info, folderinfo, this_runs_cfg
        )


def prepare_folderinfo(this_runs_results):
    # IMPORTANT NOTE
    # This function should only return None if there are issues with root_dir
    # since we handle the error in outside functions because of this (i.e. the meaning
    # behind the None should NOT change in the future!)
    if len(this_runs_results["root_dir"]) == 0:
        return
    folderinfo = {}
    folderinfo["root_dir"] = this_runs_results["root_dir"]
    # to make sure root_dir works under windows
    # (windows is okay with all dir-separators being "/", so make sure it is!)
    folderinfo["root_dir"] = folderinfo["root_dir"].replace(os.sep, "/")
    if folderinfo["root_dir"][-1] != "/":
        folderinfo["root_dir"] = folderinfo["root_dir"] + "/"
    if not os.path.exists(folderinfo["root_dir"]):
        return
    folderinfo["sctable_filename"] = this_runs_results["sctable_filename"]
    folderinfo["data_string"] = this_runs_results["data_string"]
    folderinfo["beam_string"] = this_runs_results["beam_string"]
    folderinfo["premouse_string"] = this_runs_results["premouse_string"]
    folderinfo["postmouse_string"] = this_runs_results["postmouse_string"]
    folderinfo["prerun_string"] = this_runs_results["prerun_string"]
    folderinfo["postrun_string"] = this_runs_results["postrun_string"]
    return folderinfo


# %%...............  important function getting values from tk vars  ...................


def get_results_and_cfg(results, cfg, analysis, variable_dict):
    """Before calling analysis, use .get() to extract values from tk-vars"""

    # unpack var dict
    FLOAT_VARS = variable_dict["FLOAT_VARS"]
    INT_VARS = variable_dict["INT_VARS"]
    LIST_VARS = variable_dict["LIST_VARS"]
    DICT_VARS = variable_dict["DICT_VARS"]

    # 1) We make sure to return and use for all "run"-purposes (i.e. everything
    #    that follows this local function) "this_" dicts (stuff is dangerous
    #    otherwise!)
    # 2) I implemented the usage & transformation of FLOAT_&INT_VARS the way I
    #    do here (instead of initialising as (e.g.) tk.IntVar because otherwise
    #    tkinter complains whenever an Entry is empty @ any time, just because
    #    the user is inputting their values.
    output_dicts = [{}, {}]
    for i in range(len(output_dicts)):
        if i == 0:
            input_dict = results
        elif i == 1:
            input_dict = cfg
        for key in input_dict.keys():
            # exclude mouse_num and run_num from int conversion
            # in case analysis is multi
            if analysis == "multi" and (key == "mouse_num" or key == "run_num"):
                output_dicts[i][key] = input_dict[key].get()
            elif key in FLOAT_VARS:
                # make sure that if user doesn't want to convert then we just use 0 so
                # the float() statement doesn't throw an error
                # => set the outer dict directly instead of modulating the input_dict,
                #    since that would also modify cfg which might be dangerous for
                #    future runs
                if (key == "pixel_to_mm_ratio") & (cfg["convert_to_mm"].get() is False):
                    output_dicts[i][key] = 0
                else:
                    try:
                        output_dicts[i][key] = float(input_dict[key].get())
                    except ValueError:
                        return (
                            key
                            + " value was not a decimal number, but "
                            + input_dict[key].get()
                        )
            elif key in INT_VARS:
                try:
                    output_dicts[i][key] = int(input_dict[key].get())
                except ValueError:
                    return (
                        key
                        + " value was not an integer number, but "
                        + input_dict[key].get()
                    )
            elif key in LIST_VARS:
                # if list of strings, initialise output empty list and get & append vals
                output_dicts[i][key] = []
                for list_idx in range(len(input_dict[key])):
                    output_dicts[i][key].append(input_dict[key][list_idx].get())
            elif key in DICT_VARS:
                # if dict of list of strings, initialise as empty dict and assign stuff
                # key = "angles" or other DICT_VAR
                # inner_key = "name" & "lower_ / upper_joint"
                # list_idx = idx of list of strings of inner_key
                output_dicts[i][key] = {}
                for inner_key in input_dict[key]:
                    output_dicts[i][key][inner_key] = []
                    for list_idx in range(len(input_dict[key][inner_key])):
                        output_dicts[i][key][inner_key].append(
                            input_dict[key][inner_key][list_idx].get()
                        )
            else:
                output_dicts[i][key] = input_dict[key].get()
    this_runs_results = output_dicts[0]
    this_runs_cfg = output_dicts[1]
    return this_runs_results, this_runs_cfg
